##
##                                       _ _              _
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/
##
## Alex Holehouse (Pappu Lab and Holehouse Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##
import os
import sys
import numpy
import ctypes
from . ssexceptions import SSException
from threadpoolctl import threadpool_info, threadpool_limits


##
## This is included the force numpy to use defined number of cores. For
## some of the linear algebra routines numpy will default to using as many
## cores as it can get its greedy little hands on - this function allows that
## thirst to be quenched...
##

def mkl_set_num_threads(cores):
    mkl_rt = ctypes.CDLL('libmkl_rt.so')
    mkl_get_max_threads = mkl_rt.mkl_get_max_threads()
    mkl_rt.mkl_set_num_threads(ctypes.byref(ctypes.c_int(cores)))


def _set_mkl_numpy_threads(mkl_path, num_threads):

    # Traditional UNIX-like systems will have shared objects available.
    if 'bsd' in sys.platform or 'lin' in sys.platform:
        mkl_rt = ctypes.CDLL(mkl_path)
    
    # Darwin / Apple uses `*.dylib` by default for included Intel compiler libraries.
    # Traditional UNIX-like shared objects can be created (`*.so`), but are more
    # represented in third-party libraries. This is a more dynamic way of finding
    # the MKL library and using it on a Mac that has an Intel compiler installed.
    elif sys.platform == 'darwin':
        mkl_rt = ctypes.CDLL(mkl_path)
    mkl_set_num_threads(num_threads)
    set_threads = mkl_rt.mkl_get_max_threads()
    return set_threads


def _set_openblas_numpy_threads(openblas_path, num_threads):
    # Most BLAS implementations use OpenBLAS, this section will likely be used the most
    # in production.
    openblas_lib = ctypes.cdll.LoadLibrary(openblas_path)
    openblas_lib.openblas_set_num_threads(num_threads)
    set_threads = openblas_lib.openblas_get_num_threads()
    return set_threads


def set_numpy_threads(num_threads):
    # Determine the BLAS implementation in use - e.g. MKL (Intel), OpenBLAS, etc.
    info = threadpool_info()

    # Set the threads based on the library available.
    set_threads = 0
    blas_implementation = None

    for lib in info:
        filepath = lib['filepath']
        base_filepath = os.path.basename(filepath)
        
        if 'mkl' in base_filepath:
            blas_implementation = 'mkl'
            set_threads = _set_mkl_numpy_threads(filepath, num_threads)
            break
        
        elif 'openblas' in base_filepath:
            blas_implementation = 'openblas'
            set_threads = _set_openblas_numpy_threads(filepath, num_threads)
            break

        else:
            blas_implementation = 'unknown'
    return blas_implementation, set_threads


def validate_keyword_option(keyword, allowed_vals, keyword_name, error_message=None):
    """
    Helper function that checks a passed keyword is only one of a set of possible
    valid keywords

    Parameters
    -----------

    keyword : str
        The actual passed keyword value

    allowed_vals : list of str
        A list of possible keywords

    keyword_name : str
        the name of the keyword as the user would select it in the function call

    error_message : str
        Allows the user to pass a custom error message. Default is None.


    Returns
    --------
    None

        No return value, but raises ssexceptions.SSException if ``keyword `` is not
        found in the allowed_vals list

    """


    if keyword not in allowed_vals:
        message = None
        if error_message is None:
            message = 'Keyword %s passed value [%s], but this is not valid.\nMust be one of :%s'  % (keyword_name, keyword, ", ".join(allowed_vals))
        else:
            error_type = type(error_message)
            if error_type is not str:
                raise RuntimeError('Invalid error message type: "{}". The error message must be a string.'.format(error_type))
            message = error_message[:]
        raise SSException(message)



"""
Unit and regression test for the sstrajectory module.
"""
# Import package, test suite, and other packages as needed
import os
import sys
import pytest
import ctypes
import numpy as np
from soursop import ssutils
from soursop.ssexceptions import SSException
from threadpoolctl import threadpool_info, threadpool_limits


def test_mkl_set_num_threads():
    # Determine the BLAS implementation in use - e.g. MKL (Intel), OpenBLAS, etc.
    info = threadpool_info()

    num_threads = 2
    mkl_path = None
    for lib in info:
        filepath = lib['filepath']
        base_filepath = os.path.basename(filepath)
        if 'mkl' in base_filepath:
            mkl_path = filepath[:]
            break

    if mkl_path is not None:
        # Traditional UNIX-like systems will have shared objects available.
        if 'bsd' in sys.platform or 'lin' in sys.platform:
            mkl_rt = ctypes.CDLL(mkl_path)
        
        # Darwin / Apple uses `*.dylib` by default for included Intel compiler libraries.
        # Traditional UNIX-like shared objects can be created (`*.so`), but are more
        # represented in third-party libraries. This is a more dynamic way of finding
        # the MKL library and using it on a Mac that has an Intel compiler installed.
        elif sys.platform == 'darwin':
            mkl_rt = ctypes.CDLL(mkl_path)
        ssutils.mkl_set_num_threads(num_threads)
        num_threads_set = mkl_rt.mkl_get_max_threads()
        assert num_threads_set == num_threads
    else:
        # While it may seem to be a good idea to raise an error at this point,
        # doing so puts the user in a catch-22, therefor no error should be
        # generated. A warning should be printed.
        print('Warning: No Intel BLAS implementation found.')


def test_set_numpy_threads():
    num_threads = 2
    blas_library, set_threads = ssutils.set_numpy_threads(num_threads)
    assert blas_library != 'unknown'
    assert set_threads == num_threads


def test_validate_keyword_option():
    allowed_modes = ['COM', 'CA']
    for mode in allowed_modes:
        ssutils.validate_keyword_option(mode, allowed_modes, 'mode')
        ssutils.validate_keyword_option(mode, allowed_modes, 'mode', 'Unsupported mode.')

    # Attempt a test which will fail - i.e. mismatched keyword.
    with pytest.raises(SSException):
        ssutils.validate_keyword_option('invalid_mode', allowed_modes, 'mode')

    # Attempt a test which will fail, but including a custom string error message.
    with pytest.raises(SSException):
        ssutils.validate_keyword_option('invalid_mode', allowed_modes, 'mode', 'Unsupported mode.')

    # Attempt a test which will fail, including a non-string error message.
    with pytest.raises(RuntimeError):
        ssutils.validate_keyword_option('invalid_mode', allowed_modes, 'mode', 123)



import numpy as np


def get_overlap_concentration(rg):
    """
    Function that takes the radius of gyration (Rg) in angstroms and returns 
    overlap concentration in molar units.

    The overlap concentration reflects the concentration at which a flexible
    polymer begins to 'collide' in trans with other polymers - i.e. the 
    concentration at which the chains begin to overlap.

    Parameters
    ----------

    rg : float
       Radius of gyration in Angstroms

    Return
    ------
    
    float 
        The overlap concentration in molar units

    """
    
    # Avogadro's number!
    Na = 6.023e23

    # get rg in meters (convert from Angstroms to meters)
    rg = rg*1e-10
        
    # get volume in m3
    v_m3=((4/3)*np.pi*np.power(rg, 3))
    
    # get volume in liters
    v_l = v_m3*1000
    
    # get concentration
    c = (1/(v_l))/Na

    return c



import os
import sys
import numpy as np
import pytest
import soursop
from soursop import sstrajectory
from soursop import ssprotein


GS6_FILES=['gs6.pdb','gs6.xtc']
NTL9_FILES=['ntl9.pdb','ntl9.xtc']
GROMACS_2_CHAINS=['gromacs2chains/top.pdb','gromacs2chains/traj.xtc']


test_data_dir = soursop.get_data('test_data')


@pytest.fixture(scope='session', autouse=True)
def GS6_CO(request):
    topology_path = os.path.join(test_data_dir, GS6_FILES[0])
    trajectory_path = os.path.join(test_data_dir, GS6_FILES[1])
    GS6_CO = sstrajectory.SSTrajectory(trajectory_path, topology_path)
    return GS6_CO


@pytest.fixture(scope='session', autouse=True)
def GS6_CP(request):
    topology_path = os.path.join(test_data_dir, GS6_FILES[0])
    trajectory_path = os.path.join(test_data_dir, GS6_FILES[1])
    GS6_CO = sstrajectory.SSTrajectory(trajectory_path, topology_path)
    GS6_CP = GS6_CO.proteinTrajectoryList[0]
    return GS6_CP


@pytest.fixture(scope='session', autouse=True)
def NTL9_CO(request):
    topology_path = os.path.join(test_data_dir, NTL9_FILES[0])
    trajectory_path = os.path.join(test_data_dir, NTL9_FILES[1])
    NTL9_CO = sstrajectory.SSTrajectory(trajectory_path, topology_path)
    return NTL9_CO


@pytest.fixture(scope='session', autouse=True)
def NTL9_CP(request):
    topology_path = os.path.join(test_data_dir, NTL9_FILES[0])
    trajectory_path = os.path.join(test_data_dir, NTL9_FILES[1])
    NTL9_CO = sstrajectory.SSTrajectory(trajectory_path, topology_path)
    NTL9_CP = NTL9_CO.proteinTrajectoryList[0]
    return NTL9_CP

@pytest.fixture(scope='session', autouse=True)
def GMX_2CHAINS(request):
    GMX_2CHAINS = sstrajectory.SSTrajectory("%s/%s"%(test_data_dir, GROMACS_2_CHAINS[1]), "%s/%s"%(test_data_dir, GROMACS_2_CHAINS[0]))
    return GMX_2CHAINS


# This is implemented for use in unittests for `ctanalyzer`.
# Adapted from:
# https://stackoverflow.com/questions/33508060/create-and-import-helper-functions-in-tests-without-creating-packages-in-test-di
class ProteinHelper:
    @staticmethod
    def determine_filenames(prefix, names, extension):
        expected_filenames = list()
        for name in names:
            name_parts = [prefix]
            if len(name) > 0:
                name_parts.append(name)
            savename = '%s.%s' % ('_'.join(name_parts), extension)
            expected_filenames.append(savename)
        return expected_filenames

    @staticmethod
    def lengthen_protein_trajectory(protein_traj, num_copies):
        trajectories = [protein_traj.traj for i in range(num_copies)]
        traj = protein_traj.traj.join(trajectories)
        protein = ssprotein.SSProtein(traj)
        return protein

    @staticmethod
    def validate_exported_csv_data(savename):
        assert os.path.getsize(savename) > 0  # implicit check for file existence
        with open(savename) as f:
            data = np.loadtxt(f, delimiter=',')  # numpy is operating on a buffer b/c of the temporaryfile.
            d = np.frombuffer(data).reshape(data.shape)  # convert the buffer back to an array, and compare.
            assert np.allclose(data, d)


@pytest.fixture(scope='session', autouse=True)
def cta_protein_helper():
    return ProteinHelper



"""
configs contain all global configuration.

"""
##
##                                       _ _              _ 
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/ 
##
## Alex Holehouse (Pappu Lab and Holehouse Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##
import multiprocessing as mp
import platform
import tempfile


MAXCORES = mp.cpu_count()  # Added to resolve a reference with `sstrajectory.SSTrajectory.__init__`
DEBUGGING = False


# See: https://stackoverflow.com/questions/847850/cross-platform-way-of-getting-temp-directory-in-python
TMP_DIR = tempfile.gettempdir()
if platform.system().lower() == 'darwin':
    TMP_DIR = '/tmp'






"""
sstrajectory.py

sstrajectory is the main class through which simulation trajectories are read in.

"""

##
##                                       _ _              _ 
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/ 
##
##
## Alex Holehouse (Pappu Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##

import mdtraj as md
import numpy as np
from .configs import *
from .ssdata  import ALL_VALID_RESIDUE_NAMES

from .ssprotein import SSProtein
from .ssexceptions import SSException
from . import ssutils
from . import ssio

from copy import copy


class SSTrajectory:
    """
    SSTrajectory class that holds a single simulation trajectory object. 
    

    """

    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def __init__(self, trajectory_filename=None, pdb_filename=None, TRJ=None, protein_grouping=None, pdblead=False, debug=False, extra_valid_residue_names=None):
        """
        CAMPARITraj trajectory object initializer. 

        CAMPARITraj will, by default, extract out the protein component from 
        your trajectory automatically, which lets you ask questions about the
        protein only (i.e. without salt ions getting in the way).
        
        Note that by default the mechanism by which individual proteins are
        identified is by cycling through the unique chains and determining 
        if they are protein or not. You can also provide manual grouping
        via the protein_grouping option, which lets you define which
        residues should make up an individual protein. This can be useful
        if you have multiple proteins associated with the same chain, which
        happens in CAMPARI if you have more than 26 separate chains (i.e.
        every protein after the 26th is the 'Z' chain).
               
        Parameters
        ----------        
        trajectory_filename : str
            Filename which contains the trajectory file of interest. Normally \
            this is `__traj.xtc` or `__traj.dcd`.

        pdb_filename : str
            Filename which contains the pdb file associated with the trajectory \
            of interest. Normally this is `__START.pdb`.

        TRJ : mdtraj.Trajectory
            It is sometimes useful to re-defined a trajectory and create a new SSTrajectory \
            object from that trajectory. This could be done by writing that new trajectory \
            to file, but this is extremely slow due to the I/O impact of reading/writing \
            from disk. If an mdtraj trajectory objected is passed, this is used as the \
            new trajectory from which the SSTrajectory object is constructed. 

            Default = None

        protein_grouping : list of lists of ints
            Lets you manually define protein groups to be considered independently.
        
            Default = None

        pdblead : bool
            Lets you set the PDB file (which is normally ONLY used as a topology \
            file) to be the first frame of the trajectory. This is useful when \
            the first PDB file holds some specific reference information which \
            you want to use (e.g. RMSD or Q). 

            Default = False

        debug : book
            Prints warning/help information to help debug weird stuff during initial 
            trajectory read-in. Default = False.
            

        extra_valid_residue_names : list
            By default SOURSOP identifies chains as proteins based on the a set of
            normally seen protein residue names. These are defined in soursop/ssdata,
            and are listed below:

            ['ALA','CYS','ASP','ASH','GLU','GLU','PHE','GLY','HIE','HIS','HID','HIP',
             'ILE','LEU', 'LYS','LYD','MET','ASN','PRO','GLN','ARG','SER','THR','VAL',
             'TRP','TYR','AIB', 'ABA','NVA','NLE', 'ORN', 'DAB','PTR','TPO','SEP', 'KAC', 
             'KM1', 'KM2' 'KM3', 'ACE','NME', 'FOR', 'NH2']
        
            This keyword allows the user to pass a list of ADDITIONAL residues that
            we want SOURSOP to recognize as valid residues to extract a chain as a
            protein molecule.
        """

        self.valid_residue_names = []
        self.valid_residue_names.extend(ALL_VALID_RESIDUE_NAMES)

        if extra_valid_residue_names is not None:
            try:
                self.valid_residue_names.extend(extra_valid_residue_names)
            except Exception:
                print('Unable to use the extra_valid_residue_names - this must be a list of strings')
        
        # first we decide if we're reading from file or from an existing trajectory
        if (trajectory_filename is None) and (pdb_filename is None):
            if TRJ is None:
                raise SSException('No input provided! Please provide ether a pdb and trajectory file OR a pre-formed traj object')
                
            # note the [:] means this is a COPY!
            self.traj = TRJ[:]
        else:
            if (trajectory_filename is None):
                raise SSException('No trajectory file provided!')
            if (pdb_filename is None):
                raise SSException('No PDB file provided!')

            # read in the raw trajectory
            self.traj = self.__readTrajectory(trajectory_filename, pdb_filename, pdblead)


        # Next, having read in the trajectory we parse out into proteins
        # extract a list of protein trajectories where each protein is assumed
        # to be in its own chain
        if protein_grouping == None:
            self.proteinTrajectoryList = self.__get_proteins(self.traj, debug)        
        else:
            self.proteinTrajectoryList = self.__get_proteins_by_residue(self.traj, protein_grouping, debug)

        self.__single_protein_traj = self.__get_all_proteins(self.traj)

        
        self.num_proteins = len(self.proteinTrajectoryList)
        self.n_frames = len(self.traj)


    def  __repr__(self):
        return "SSTrajectory (%s): %i proteins and %i frames" % (hex(id(self)), self.num_proteins, self.n_frames)


    def __len__(self):
        # Edited to mimic the behavior of `mdtraj` trajectory objects.
        # Originally: `return (self.num_proteins, self.n_frames)`
        return self.n_frames


    def length(self):
        # Implemented a method that encapsulates the data output by the original `__len__` method.
        return (self.num_proteins, self.n_frames)


    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def __readTrajectory(self, trajectory_filename, pdb_filename, pdblead):
        """
        Internal function which parses and reads in a CAMPARI trajectory

        Read a trajectory file. This was separated out into its own
        function in case we want to add additional sanity checks during
        the file loading.

        Notably older versions of CAMPARI mess up the unitcell length
        vectors, so will cause problems, but you can get around this by
        having GROMACS rebuild the trajectory. If this happens an error
        pops up but instructions on how to use GROMACS to fix it are
        presented.

        Parameters
        -----------
        trajectory_filename : str
            Filename which contains the trajectory file of interest. File type
            is automatically detected and dealt with mdtraj' 'load' command
            (i.e. md.load(filename, top=pdb_filename))

        pdb_filename : str
            Filename which contains the pdb file associated with the trajectory
            of interest. This defines the topology of the system and must match
            the trajectory in terms of number of atomas


        pdblead : bool
            Also extract the coordinates from the PDB file and append it to 
            the front of the trajectory. This is useful if you are starting
            an analysis where that first structure should be a reference frame
            but it's not actually included in the trajectory file.

        Returns
        --------
        mdtraj.traj 
            Returns an mdtraj trajectory object

        """

        # straight up read the trajectory first using mdtraj's awesome
        # trajectory reading facility
        traj =  md.load(trajectory_filename, top=pdb_filename)
                    
        # check unit cell lengths
        try:
            uc_lengths = traj.unitcell_lengths[0]

            # this is s custom warning for a specific edge-case we encounter a lot
            if (uc_lengths[0] == 0 or uc_lengths[1] == 0 or uc_lengths[2] == 0):
                ssio.warning_message("Trajectory file unit cell lengths are zero for at least one dimension. This is a probably a bug with an FRC generated __START.pdb file, because in the old version of CAMPARI used to do grid based FRC calculations the unit cell dimensions are not written correctly. This may cause issues but we're going to assume everything is OK for now. Check the validity of any analysis output. If you're worried, you can use the following workaround.\n\n:::: WORK AROUNDS ::::\nSimply run\n\ntrjconv -f __traj.xtc -s __START.pdb -box a b c -o frc.xtc \n\nAn then \n\ntrjconv -f frc.xtc -s __START.pdb -box a b c -o start.pdb -dump 0\n\n\nHere\n-f n__traj.xtc   : defines the trajectory file\n-s __start.pdb   : defines the pdb file used to parse the topology\n-box a b c       : defines the box lengths **in nanometers**\n-o frc.xtc       : is the name of the new trajectory file with updated box lengths\nSelect 0 (system) when asked to 'Select group for output'.The second step creates the equivalent PDB file with the header-line correctly defining the box unit cell lengths and angles. These two new files should then be used for analysis.\n\nAs an example, if my FRC simulation had a sphere radius of 100 angstroms then my correction command would look something like \n\ntrjconv -f __traj.xtc -s __START.pdb -box 20 20 20 -o frc.xtc\ntrjconv -f frc.xtc -s __START.pdb -box 20 20 20 -o start.pdb -dump 0")

        except TypeError:
            ssio.warning_message("Warning: UnitCell lengths were not provided... This may cause issues but we're going to assume everything is OK for now...")
        
        # if pdbLead is true then load the pdb_filename as a trajectory
        # and then add it to the front (the PDB file is its own topology
        # file so no need to specificy the top= file here!
        if pdblead:
            pdbtraj = md.load(pdb_filename)
            traj = pdbtraj+traj


            # having added the PDB file now check all the unit-cells match up!
            try:
                uc_lengths_1 = traj.unitcell_lengths[0]
                uc_lengths_2 = traj.unitcell_lengths[1]
            
                if (uc_lengths_1[0] != uc_lengths_2[0]) or (uc_lengths_1[2] != uc_lengths_2[2]) or (uc_lengths_1[2] != uc_lengths_2[2]):
                    ssio.warning_message('........................\nWARNING:\nThe unit cell dimensions of the PDB file and trajectory file did not match, specifically\nPDB file = [ %s ]\nXTC file = [ %s ]\nThis may cause issues if native MDTraj untilities are used (and potentially for SOURSOP utilities that are based on these. It is not necessarily an issue, but PLEASE sanity check your outcome. To be save we reeommend editing the PDB-file untilcell dimenions to match.')

            except TypeError:
                ssio.warning_message("Warning: UnitCell lengths were not provided... This may cause issues but we're going to assume everything is OK for now...")
                                                                           
        return traj



    def __get_all_proteins(self, trajectory):
        """
        Internal function that builds a single trajectory which contains all protein
        residues. 

        Parameters
        -----------
        trajectory : mdtraj.Trajectory
            An already parsed trajectory object (i.e. checked for CAMPARI-
            relevant defects such as unitcell issues etc)

        Returns
        ----------
        ssprotein.SSProtein
            Returns a single SSProtein object
        
        """

        

        # extract full system topology
        topology = trajectory.topology

        protein_atoms = []
        
        # for each chain in this toplogy determine if the 
        # first residue is protein or not. If it's protein we parse it if 
        # not it gets skipped
        for chain in topology.chains:

            # if the first residue in the chain is protein
            # note that a formic acid cap ('FOR') is not recognized as protein
            # so we include an edgecase here for that
            if chain.residue(0).name in self.valid_residue_names:

                # intialize an empty list of atoms
                local_atoms = []

                # get every atom in this chain
                for atom in chain.atoms:
                    protein_atoms.append(atom.index)

                protein_atoms.extend(local_atoms)

        return SSProtein(trajectory.atom_slice(protein_atoms))
        

    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def __get_proteins(self, trajectory, debug):
        """
        Internal function that takes an MDTraj trajectory and returns a list of mdtraj 
        trajectory objects corresponding to each protein in the system, ASSUMING that 
        each protein is in its own chain.
        
        The way this works is to cycle through each chain, identify if that chain
        contains protein or not, and if it does grab all the atoms in that chain
        and perform an atomslice using those atoms on the main trajectory.
        

        Parameters
        -----------
        trajectory : mdtraj.Trajectory
            An already parsed trajectory object (i.e. checked for CAMPARI-
            relevant defects such as unitcell issues etc)

        Returns
        ---------
        list
            Returns a proteinTrajectoryList - a list with one or more SSProtein
            objects in it
        
        """

        # extract full system topology
        topology = trajectory.topology



        chainAtoms = []
        
        # for each chain in this toplogy determine if the 
        # first residue is protein or not. If it's protein we parse it if 
        # not it gets skipped
        for chain in topology.chains:

            # if the first residue in the chain is protein
            # note that a formic acid cap ('FOR') is not recognized as protein
            # so we include an edgecase here for that
            if chain.residue(0).name in self.valid_residue_names:

                # intialize an empty list of atoms
                local_atoms = []

                # get every atom in this chain
                for atom in chain.atoms:
                    local_atoms.append(atom.index)

                chainAtoms.append(local_atoms)

            else:
                if debug:
                    ssio.debug_message('Skipping residue %s from %s' %(chain.residue(0).name, chain))

        # for each protein chain that we have atomic indices
        # for (hopefully all of them!) cycle through and create
        # sub-trajectories
        proteinTrajectoryList = []
        for local_chain_atoms in chainAtoms:

            # generate a trajectory composed of *JUST* the
            # $chain atoms. The PT object created now is self
            # consistent and contains an associated and fully
            # correct .topology object (NOTE this fixes a 
            # previous bug in CAMPARITraj 0.1.4)
            PT = trajectory.atom_slice(local_chain_atoms)

            # WA
            # gets the resid offset in a way that is ensures internal
            # consistency for the SSProtein object
            first_resid = PT.topology.chain(0).residue(0).index

            if first_resid != 0:
                raise SSException('After extracting a protein subtrajectory, the first resid is not 0. This may reflect a bug, or you may not be using MDTraj 1.9.5')
                
            # add that SSProtein to the ever-growing proteinTrajectory list
            proteinTrajectoryList.append(SSProtein(PT))

        if len(proteinTrajectoryList) == 0:
            ssio.warning_message('No protein chains found in the trajectory')

        return proteinTrajectoryList




    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def __get_proteins_by_residue(self, trajectory, residue_grouping, debug):
        """
        Internal function which returns a list of mdtraj trajectory objects corresponding 
        to each protein where we *explicitly* define the residues in each protein.

        Unlike the `__get_proteins()` function, which doesn't require any manual
        input in identifying the proteins, here we provide a list of groups, where 
        each group is the set of residues associated with a protein.

        The way this works is to cycle through each group, and for each residue in 
        each group grabs all the atoms and uses these to carry out an atomslice 
        on the full trajectory. 

        Parameters
        -----------
        trajectory : mdtraj.Trajectory
            An already parsed trajectory object (i.e. checked for CAMPARI-
            relevant defects such as unitcell issues etc)

        residue_grouping : list of list of integers
            Must be a list containing one or more lists, where each internal list
            contains a set of monotonically increasing residues (which correspond
            to the full protein trajectory). In other words, each sub-list defines
            a single protein. The integer indexing here - importantly - uses the 
            CAMPARITraj internal residue indexing, meaning that indexing begins at 0 from
            the first residue in the PDB file.

        Returns
        ---------
        list
            Returns a proteinTrajectoryList - a list with one or more SSProtein
            objects in it
        

        """
        
        group_atoms = []

        # extract full system topology
        topology = trajectory.topology
        
        # for each chain in this toplogy determine if the 
        # first residue is protein or not        
        for group in residue_grouping:


            # build a string of the resids
            res_string = ''
            for r in group:
                res_string = res_string + " %i" % (int(r)) 
            
            # select atoms based on the resid string
            local_atoms = topology.select('resid %s' %(res_string))
            
            if len(local_atoms) == 0:
                ssio.warning_message('In residue group [%s ...] no residues in the trajectory were found...' %(str(group)[0:8]))

            else:
                group_atoms.append(local_atoms)

        # for each protein group that we have atomic indices
        # for cycle through and create sub-trajectories
        proteinTrajectoryList=[]
        
        # cycle through each group of atoms as was defined by the residue_grouping
        # indices

        for local_group_atoms in group_atoms:

            # generate a trajectory composed of *JUST* the
            # $chain atoms. The PT object created now is self
            # consistent and contains an associated and fully
            # correct .topology object (NOTE this fixes a 
            # previous bug in CAMPARITraj 0.1.4)
            PT = trajectory.atom_slice(local_group_atoms)       
            
            # gets the resid offset in a way that is ensures internal
            # consistency for the SSProtein object
            resid_offset = PT.topology.chain(0).residue(0).index

            if resid_offset != 0:
                raise SSException('After extracting a protein subtrajectory, the first resid is not 0. This may reflect a bug, or you may not be using MDTraj 1.9.5')
                
            proteinTrajectoryList.append(SSProtein(PT))

        if len(proteinTrajectoryList) == 0:
            ssio.warning_message('No protein chains found in the trajectory')


        return proteinTrajectoryList


    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def get_overall_radius_of_gyration(self):
        """
        Function which returns the per-frame OVERALL radius of gyration for every 
        protein residue in the trajectory. For systems with multiple protein chains, 
        all chains are combined together. For systems with a single protein chain,
        this function offers no advantage over interacting directly with the
        SSProtein object in the .proteinTrajectoryList.

        Parameters
        -------------
        None

        Returns
        ----------
        np.ndarray 
            Returns a numpy array with per-frame instantaneous radius of gyration
        
        """

        return self.__single_protein_traj.get_radius_of_gyration()

    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def get_overall_asphericity(self):
        """
        Function which returns the per-frame OVERALL asphericity for every
        protein residue in the trajectory. For systems with multiple protein chains, 
        all chains are combined together. For systems with a single protein chain,
        this function offers no advantage over interacting directly with the
        SSProtein object in the .proteinTrajectoryList.

        Parameters
        -------------
        None

        Returns
        ----------
        np.ndarray 
            Returns a numpy array with per-frame instantaneous asphericity
        """

        return self.__single_protein_traj.get_asphericity()

    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def get_overall_hydrodynamic_radius(self):
        """
        Function which returns the per-frame OVERALL hydrodynamic radius for every
        protein residue in the trajectory. For systems with multiple protein chains, 
        all chains are combined together. For systems with a single protein chain,
        this function offers no advantage over interacting directly with the
        SSProtein object in the .proteinTrajectoryList.

        Parameters
        -------------
        None

        Returns
        ----------
        np.ndarray 
            Returns a numpy array with per-frame instantaneous asphericity

        """

        return self.__single_protein_traj.get_hydrodynamic_radius()


    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def get_interchain_distance_map(self, proteinID1, proteinID2, mode='CA'):
        """        
        Function which returns two matrices with the mean and standard deviation distances
        between the residues in resID1 from proteinID1 and resID2 from proteinID2

        This computes the (full) intramolecular distance map, where the "distancemap"
        function computes the intermolecular distance map.

        Specifically, this allows the user to define two distinct chains (i.e. an "interchain"
        distance map).

        Obviously this only makes sense if your system has two separate protein objects
        defined, but in principle the output from:

            `intra_chain_distance_Map(0,0)`

        would be the same as:
        
            `proteinTrajectoryList[0].get_distance_map()`

        This is actually a useful sanity check!
        
        Parameters
        ----------

        proteinID1 : int
            The ID of the first protein of the two being considered, where the ID is the proteins position in the
            `self.proteinTrajectoryList` list. 

        proteinID2 : int
            The ID of the second protein of the two being considered, where the ID is the proteins position in the
            `self.proteinTrajectoryList` list

        mode : ['CA','COM']
            String, must be one of either 'CA' or 'COM'.
            - 'CA' = alpha carbon.
            - 'COM' = center of mass (associated withe the residue).
            Default = 'CA'.


        Returns
        -------
        tuple : tuple containing `distanceMap` and `STDMap`

        distanceMap : numpy matrix
            Is an [n x m] matrix where n and m are the number of proteinID1 residues
            and proteinID2 residues. Each position in the matrix corresponds to the
            mean distance between those two residues over the course of the simulation.

        stdMap : numpy matrix
            Is an [n x m] matrix where n and m are the number of proteinID1 residues
            and proteinID2 residues. Each position in the matrix corresponds to the
            standard devaiation associated with the distances between those two
            residues.

        
        """

        ssutils.validate_keyword_option(mode, ['CA', 'COM'], 'mode')
        
        # get SSProtein objects for the two IDs passed (could be the same)
        P1 = self.proteinTrajectoryList[proteinID1]        
        P2 = self.proteinTrajectoryList[proteinID2]

        # create the empty distance maps
        p1_residues = P1.resid_with_CA
        p2_residues = P2.resid_with_CA
        map_shape   = (len(p1_residues), len(p2_residues))
        distanceMap = np.zeros(map_shape)
        stdMap      = np.zeros(map_shape)

        for r1 in p1_residues:
            if mode == 'COM':
                COM_1 = P1.get_residue_COM(r1)
            else:
                COM_1 = P1.get_residue_COM(r1, atom_name='CA')

            p1_index = copy(r1)
            if P1.ncap:
                p1_index = r1 - 1

            for r2 in p2_residues:

                if mode == 'COM':
                    COM_2 = P2.get_residue_COM(r2)
                else:
                    COM_2 = P2.get_residue_COM(r2, atom_name='CA')

                p2_index = copy(r2)
                if P2.ncap:
                    p2_index = r2 - 1
                
                # compute distance...
                d = 10*np.sqrt(np.square(np.transpose(COM_1)[0] - np.transpose(COM_2)[0]) + np.square(np.transpose(COM_1)[1] - np.transpose(COM_2)[1])+np.square(np.transpose(COM_1)[2] - np.transpose(COM_2)[2]))

                
                distanceMap[p1_index, p2_index] =  np.mean(d, 0)
                stdMap[p1_index, p2_index]    =  np.std(d, 0)
                
        return (distanceMap, stdMap)


    #oxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxoxoxoxoxoxoxoxoxoxooxoxo
    #
    #
    def get_interchain_distance(self, proteinID1, proteinID2, R1, R2, A1='CA', A2='CA', mode='atom', periodic=False):
        """
        Function which returns the distance between two specific atoms on two residues, or between 
        two residues based on mdtraj' atomselection mode rules (discussed below). Required input are protein
        ID selectors and the resid being used. Resids should be used as would be normally used for the SSProtein
        objects associated with proteinID1 and proteinID2.

        For inter-atomic distances, atoms are selected from the passed residue and their 'name' field from the topology 
        selection language (e.g. "CA", "CB" "NH" etc). By default CA atoms are used, but one can define
        any residue of interest. We do not perform any sanity checking on the atom name - this gets 
        really hard - so have an explicit try/except block which will warn you that you've probably 
        selected an illegal atom name from the residues.

        For inter-residue distances the associated rules are defined by the 'mode' selector. By default
        mode is set to 'atom', which means the variables A1 and A2 are used (with CA as default) to define
        inter-residue distance. However, if one of the other modes are used the A1/A2 parameters are ignored
        and alternative rules for computing inter-residue distance are used. These modes are detailed below.

        Distance is returned in Angstroms.

        Parameters
        ----------

        proteinID1 : int
            Index of the first protein of interest

        proteinID2 : int
            Index of the second protein of interest

        R1 : int
            Residue index of first residue

        R2 : int
            Residue index of second residue

        A1 : str
            Atom name of the atom in R1 we're looking at. Default = 'CA'

        A2 : str
            Atom name of the atom in R2 we're looking at. Default='CA'

        mode : str
            Mode allows the user to define different modes for computing atomic distance.

            The default is 'atom' whereby a pair of atoms (A1 and A2) are provided. Other options
            are detailed below and are identical to those offered by mdtraj in compute_contacts.

            Note that if modes other than 'atom' are used the A1 and A2 options are ignored.

            + `'ca'` - same as setting `'atom'` and `A1='CA'` and `A2='CA'`, this uses the C-alpha atoms.
            + `'closest'` - closest atom associated with each of the residues, i.e. the is the point \
                            of closest approach between the two residues.
            + `'closest-heavy'` - same as `'closest'`, except only non-hydrogen atoms are considered.
            + `'sidechain'` - closest atom where that atom is in the sidechain.
            + `'sidechain-heavy'` - closest atom where that atom is in the sidechain and is heavy.

        periodic : Bool
            Flag which if distances mode is passed as anything other than 'atom' then this determines
            if the minimum image convention should be used. Note that this is only available if pdb
            crystal dimensions are provided, and in general it's better to set this to false and 
            center the molecule first. Default = False.

        Returns
        -----------
        np.array
            Returns a 1D numpy array with the distance-per-frame betwee the specified residues

        """

        # check mode keyword is valid
        allowed_modes = [ 'atom', 'ca', 'closest', 'closest-heavy', 'sidechain', 'sidechain-heavy' ]
        if mode not in allowed_modes:
            raise SSException("Provided mode keyword must be one of 'atom', 'ca', 'closest', 'closest-heavy', 'sidechain', or 'sidechain-heavy'. Provided keyword was [%s]" % (mode))

        # get SSProtein objects for the two IDs passed (could be the same)        
        try:
            P1 = self.proteinTrajectoryList[proteinID1]
            P2 = self.proteinTrajectoryList[proteinID2]

        except IndexError as e:
            raise SSException('In get_interchain_distance(): When selecting protein indices %i and %i at least one of these was out of range (indices are from 0...%i)' % (proteinID1, proteinID2, len(self.proteinTrajectoryList)-1))
            

        # next build a new trajectory that contains ONLY the two residues selected
        local_atoms1 = P1.topology.select('resid %i' % (R1))
        local_atoms2 = P2.topology.select('resid %i' % (R2))

        if len(local_atoms1) == 0:
            raise SSException("In get_interchain_distance(): When selecting resid %i from proteinID1 found no atoms" %(R1))

        if len(local_atoms2) == 0:
            raise SSException("In get_interchain_distance(): When selecting resid %i from proteinID2 found no atoms" %(R2))
            
        subtraj_p1 = P1.traj.atom_slice(local_atoms1)
        subtraj_p2 = P2.traj.atom_slice(local_atoms2)
        
        # this is now a subtrajectory which in principle contains just two residues. We can check
        # this to ensure that the trajectory has exactly 2 residues
        full_subtraj = subtraj_p1.stack(subtraj_p2)
        full_subtraj_residues = [i for i in full_subtraj.topology.residues]
        if len(full_subtraj_residues) != 2:
            raise SSException("In get_interchain_distance(): When passed in two residues (R1=%i, R2=%i) in proteins %i and %i found multiple residues (%i)...these resids could not be found " %(R1, R2, proteinID1, proteinID2, len(full_subtraj_residues)))

        
        # if we're looking at a specific pair of atoms (note we use resid 0 and 1 because we KNOW this trajectory only has 2 residues and we know R1 is 0 and R2 is 1
        if mode == 'atom':

            atom1 = full_subtraj.topology.select('resid 0 and name "%s"' % A1)
            if len(atom1) != 1:
                raise SSException("In get_interchain_distance() when selecting atom %s from residue %i in protein %i no atoms were found " % (A1, R1,  proteinID1))                

            COM_1 = md.compute_center_of_mass(full_subtraj.atom_slice(atom1))

            atom2 = full_subtraj.topology.select('resid 1 and name "%s"' % A2)
            if len(atom2) != 1:
                raise SSException("In get_interchain_distance() when selecting atom %s from residue %i in protein %i no atoms were found " % (A2, R2,  proteinID2))                

            COM_2 = md.compute_center_of_mass(full_subtraj.atom_slice(atom2))
            
            # finally compute distances
            distances = 10*np.sqrt(np.square(np.transpose(COM_1)[0] - np.transpose(COM_2)[0]) + np.square(np.transpose(COM_1)[1] - np.transpose(COM_2)[1])+np.square(np.transpose(COM_1)[2] - np.transpose(COM_2)[2]))

        else:

            # TODO: Documentation missing!
            # use the compute_contacts() function from mdtraj, multiplying by 10 because this will
            # by default give you numbers that... 
            distances = 10*md.compute_contacts(full_subtraj, [[0, 1]], scheme=mode, periodic=periodic)[0].ravel()

        return distances
      




##
##
##                                       _ _              _ 
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/ 
##
## Alex Holehouse (Pappu Lab and Holehouse Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##

import mdtraj as md
import numpy as np
import scipy
from .ssexceptions import SSWarning

### SSPRE contains all the functionality associated with calculating
### PRE profiles.
###
###
###
###


original_K = 1.2300e-32       # K constant in cm6*s-2
K_IN_NM6   = original_K*1e42  # K constant in nm6 s-2
#W_H        = 267530000        # Proton Larmor frequency
#W_H_SQUARED = W_H * W_H 

class SSPRE:
    """
    
    SSPRE is a class that takes a SSProtein objecs and can perform PRE-related calculations.
       
    This is, in many ways, a purely functional class, but given it's very specific goal
    it is separated out into its own class in the interest of more robust modularity. 

    There are a number of internal functions, but the only public facing function is the 
    generate_PRE_profile function below. This gives both the intensity ratio profile and 
    the transverse relaxation rate (Gamma_2) profiles. For more detail on the calculation
    of these profiles see the help associated with generate_PRE_profile function

    For more information on caculation of PREs using method see the Supplementary information 
    in the following two papers.

    Meng, W., Lyle, N., Luan, B., Raleigh, D.P., and Pappu, R.V. (2013). Experiments and simulations 
    show how long-range contacts can form in expanded unfolded proteins with negligible secondary structure. 
    Proc. Natl. Acad. Sci. U. S. A. 110, 2123-2128.

    Das, R.K., Huang, Y., Phillips, A.H., Kriwacki, R.W., and Pappu, R.V. (2016). Cryptic sequence 
    features within the disordered protein p27Kip1 regulate cell cycle signaling. 
    Proc. Natl. Acad. Sci. U. S. A. 113, 5616- 5621.


    """

    # ........................................................................
    #                 
    def __init__(self, SSProteinObject, tau_c, t_delay, R_2D, W_H):
        """
        Initialization function for creating a SSPRE object. The resulting object can
        then be used to calculate PRE profiles from the underlying ensemble. This
        calculation is extremely fast.

        SSProteinObject : SSProtein-derived object
            SSProtein object extracted from a SSTrajectory objected. The SSProtein object is the main object
            that most protein-based analysis is performed over in SOURSOP.

        tau_c : float
            tau_c is the effective correlation time, measured in nanoseconds, which is typically between
            1 and 30.

        t_delay : float
            Total duration of the INEPT delays from the PRE experiment, as measured in ms. This will
        depend on the pulse sequence used, but is typically around 1-30 ms for HSQC.

        R_2D : float
            Is the transverse relaxation rate of the backbone amide protons in the diamagnetic form
            of the protein, measured in Herz (i.e. 'per second'). A value of around 10 might be expected.

        W_H : float
            Is the proton Larmor frequency, which is typically the "MHz" value associated with the magnet,
            given in Hz. For examle, a 600 MHz magnet would use the value 600000000.

            (Note that the proton Larmor frequency at 1 Tesla = 267530000 per second per Tesla).

        """
                
        # NOTE: NO CHANGES SHOULD BE MADE TO THE SSPO by the object - this should be treated 
        # as a read only object (no such explicit control in Python)
        self.SSPO = SSProteinObject        
        
        # set the INEPT delay value and the backbone amide transverse relaxation rate which
        # is used explicitly later
        self.t_delay = float(t_delay)   # in ms - INDEPT delay
        self.R_2D = float(R_2D)         # in Hz - backbone amide transverse relaxation rate
        self.tau_c = tau_c              # in ns - effecive correlation time
        self.W_H   = W_H                # in Hz - Proton Larmor frequency in the magnet

        # ------------------------------------------
        # sanity checks to warn if any of the input values seem dratsically wrong. NOTE that these won't block the analysis but will
        # throw up errors
        if self.R_2D < 0.01 or self.R_2D > 100:
            CTWarning("WARNING: The value of R_2D (bacbone amide transverese relaxation rate) is far from the normal expected value of ~10 (R_2D = %4.4e) - recal this value is in units of Herz" %(self.R_2D))

        if self.t_delay < 0.01 or self.t_delay > 100:
            CTWarning("WARNING: The value of t_delay (INEPT delay) is far from the normal expected value of ~15 (t_delay = %4.4e) - recal this is in units of ms" %(self.t_delay))

        if tau_c < 0.01 or tau_c > 100:
            CTWarning("WARNING: The value of tau_c (effective correlation time) is far from the normal expected value of ~5 (t_delay = %4.4e) - recal this is in units of ns" %(self.tau_c))

        # if Larmor frequency less than 100 MhZ or above 2 GHz assume something is wrong
        if W_H < 50000000 or tau_c > 2000000000:
            CTWarning("WARNING: The value of W_h (proton Larmor frequency) is far from the normal expected value of ~600 000 000 - recal this value should be provided in Herz" %(self.W_H))

        # # convert tau_c to seconds and calculate tau_c squared
        tau_c = float(tau_c)/1000000000     # tau c in seconds
        tau_c_squared = tau_c * tau_c       # 
        
        # compute the prefactor term which will be used when computing the PRE dependent relaxation profile
        # by the generate_PRE_profile function
        W_H_SQUARED = W_H*W_H
        PREFACTOR = (3 * tau_c)/(1 + W_H_SQUARED * tau_c_squared)
        PREFACTOR = (4*tau_c + PREFACTOR)
        self.PREFACTOR = PREFACTOR * K_IN_NM6


    # ........................................................................
    #                 
    def __repr__(self):
        """


        """
        return "["+hex(id(self)) + "]: SSPRE OBJ - (R_2D = %3.2f Hz, t_delay = %3.2f ms, tau_c = %3.2f ns, H1 Larmor = %3.3e Hz)" % (self.R_2D, self.t_delay, self.tau_c, self.W_h)


    # ........................................................................
    #                 
    def generate_PRE_profile(self, label_position, spin_label_atom='CB', target_relaxation_atom='N'):
        """
        Construct a PRE intensity profile and gamma profile based on a nitroxide spin label being placed at the label position
        position on the $spin_label_atom atom. By default this is the CB and the PRE distance to be used to asses relaxation
        comes from the CB-backbone N distances (as used in work by Meng & Lyle [1] and Das [2]).

        The PRE profile describes the intensity ratio (I_paramagnetic / I_diamagnetic), and typically varies between 0 and 1.
        When the intensity ratio ~0 the spin label dominantes relaxation and suggests the residue in question is near the
        spin label. When the intensity ratio is ~1 the relaxation is primarily through non-spin label mediated mechanisms
        suggesting the label and the residue are far apart.

        The gamma profile describes the spin-label induced amide proton relaxation rate in units of per-second, providing
        another observable that can be directly compared with experiment. 

        This function is extremely fast (sub 10 seconds on a ~6000 frame ensemble).

        Returns
        -------
        Returns a 2 place tuple - tuple position 0 is the PRE intensity profile and tuple position 1 is the PRE H1 relaxatation
        profile.

        References
        ----------

        [1] Meng, W., Lyle, N., Luan, B., Raleigh, D.P., and Pappu, R.V. (2013). Experiments and simulations show how long-range 
        contacts can form in expanded unfolded proteins with negligible secondary structure. 
        Proc. Natl. Acad. Sci. U. S. A. 110, 2123-2128.
        [2] Das, R.K., Huang, Y., Phillips, A.H., Kriwacki, R.W., and Pappu, R.V. (2016). Cryptic sequence features within the 
        disordered protein p27Kip1 regulate cell cycle signaling. Proc. Natl. Acad. Sci. U. S. A. 113, 5616- 5621.
               
        Parameters
        ----------

        label_position : int
            Position in the sequence at which the spin-label is located. Should ideally contain a CB atom (i.e. not be be glycine),
            else the label atom must be set to 'CA' see below.

        spin_label_atom : str, default='CB'
            Name of the atom upon which the spin label is located. Should really be CB but may be changed if a residue lacks a CB atom
            (e.g. a glycine is in the place of the Cys nitroxide spin labeled residue). Ideally

        target_relaxation_atom : str, default='N'
            Name of the atom where relaxation is being performed. This should be 'N' (backbone amide) as that's how this approach is
            parameterized - highly recommended that this isn't changed. If it is changed the method will look for an atom of this name
            in every residue. Again, it is STRONGLY recommended this isn't changed.
        """
        
        # get index value of all residues 
        #residue_list = self.SSPO.get_residue_index_list()        

        tmp = list(self.SSPO._SSProtein__CA_residue_atom.keys())
        residue_list = sorted(tmp)

        # first calculate mean rij distance for pair-residue distances
        gamma = []

        # finally for each residue calculate the r^6 distances associated with each frame and for EACH FRAME calculate 
        # the PREFACTOR / r^6 value and then take the mean. THIS gives a different answer to if you take the mean distance
        # and calculate the PREFACTOR/<R^6> value because there is a non-linear mapping between relaxation and distance so 
        # it's important the former method is used (i.e. only average at the end). This calculates the gamma coefficient for
        # each residue, which measures relaxation 
        for idx in residue_list:                        
            r_6_nm = np.power((0.1*self.SSPO.get_inter_residue_atomic_distance(label_position, idx, A1=spin_label_atom, A2=target_relaxation_atom),6)
            gamma.append(np.mean(self.PREFACTOR/r_6_nm))
        
        # convert the t_delay from ms to seconds
        t_delay_in_seconds = self.t_delay/1000 
        
        # for each gamma compute the intensity ration 
        profile = []
        for g in gamma:
            profile.append((self.R_2D * np.exp(-g*t_delay_in_seconds)) / (self.R_2D + g))        

        return (profile, gamma)





"""
Unit and regression test for the soursop package.
"""

# Import package, test suite, and other packages as needed
import numpy as np
import soursop
import pytest
import sys
import random
import itertools
from copy import deepcopy
from soursop import sstrajectory, ssprotein
from soursop.ssexceptions import SSException
from soursop.configs import DEBUGGING


def test_contacts_special(NTL9_CP):

    with pytest.raises(SSException) as error:
        a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='sidechain-invalid')

def test_code_coverage(NTL9_CP):

    NTL9_CP.print_residues()

    # property
    a = NTL9_CP.residue_index_list
    a = NTL9_CP.n_residues
    a = NTL9_CP.n_frames



    a = NTL9_CP.get_amino_acid_sequence()
    a = NTL9_CP.get_CA_index(10)
    a = NTL9_CP.get_multiple_CA_index()
    a = NTL9_CP.get_multiple_CA_index([3,4,5])


    a = NTL9_CP.calculate_all_CA_distances(10)
    a = NTL9_CP.calculate_all_CA_distances(10, stride=2)
    a = NTL9_CP.calculate_all_CA_distances(10, stride=2)
    a = NTL9_CP.calculate_all_CA_distances(10, stride=2, mode='COM')
    a = NTL9_CP.calculate_all_CA_distances(10, stride=2, mode='COM', only_C_terminal_residues=False)

    a = NTL9_CP.get_distance_map()
    a = NTL9_CP.get_distance_map(verbose=True)
    a = NTL9_CP.get_distance_map(verbose=True, mode='COM')
    a = NTL9_CP.get_distance_map(verbose=True, mode='COM', RMS=True)

    a = NTL9_CP.get_polymer_scaled_distance_map()
    a = NTL9_CP.get_polymer_scaled_distance_map(nu=0.54,A0=6)
    a = NTL9_CP.get_polymer_scaled_distance_map(nu=0.54,A0=6,min_separation=20)
    a = NTL9_CP.get_polymer_scaled_distance_map(nu=0.54,A0=6,min_separation=20, mode='signed-fractional-change')
    a = NTL9_CP.get_polymer_scaled_distance_map(nu=0.54,A0=6,min_separation=20, mode='signed-absolute-change')
    a = NTL9_CP.get_polymer_scaled_distance_map(nu=0.54,A0=6,min_separation=20, mode='scaled')

    a = NTL9_CP.get_local_heterogeneity(stride=1)
    a = NTL9_CP.get_local_heterogeneity(fragment_size=5, stride=1)
    a = NTL9_CP.get_local_heterogeneity(fragment_size=20, stride=2)
    a = NTL9_CP.get_local_heterogeneity(fragment_size=20, stride=1)
    a = NTL9_CP.get_D_vector(stride=1)

    a = NTL9_CP.get_RMSD(1)
    a = NTL9_CP.get_RMSD(1,frame2=3)
    a = NTL9_CP.get_RMSD(1,frame2=3, stride=2)
    a = NTL9_CP.get_RMSD(1,frame2=3, stride=2, region=[10,20])
    a = NTL9_CP.get_RMSD(1,frame2=3, stride=2, backbone=False)


    a = NTL9_CP.get_Q()
    a = NTL9_CP.get_Q(stride=2)
    a = NTL9_CP.get_Q(stride=2, protein_average=False)
    a = NTL9_CP.get_Q(stride=2, protein_average=False, region=[10,20])


    a = NTL9_CP.get_contact_map()
    a = NTL9_CP.get_contact_map(distance_thresh=2)
    a = NTL9_CP.get_contact_map(distance_thresh=2, stride=2)
    a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='ca')
    a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='closest-heavy')
    a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='closest')
    a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='sidechain')

    # This test fails. Will revisit.
    # a = NTL9_CP.get_contact_map(distance_thresh=2, stride=1, mode='sidechain-heavy')

    a = NTL9_CP.get_clusters(stride=1)
    a = NTL9_CP.get_clusters(stride=1, n_clusters=3)
    a = NTL9_CP.get_clusters(stride=1, n_clusters=3, backbone=False)


    a = NTL9_CP.get_inter_residue_COM_distance(10,30)
    a = NTL9_CP.get_inter_residue_COM_distance(10,30, stride=2)


    a = NTL9_CP.get_inter_residue_COM_vector(10,30)
    a = NTL9_CP.get_inter_residue_COM_vector(10,30, stride=2)


    a = NTL9_CP.get_inter_residue_atomic_distance(2,10)
    a = NTL9_CP.get_inter_residue_atomic_distance(2,10, A1='CB')
    a = NTL9_CP.get_angle_decay()


def test_get_radius_of_gyration(GS6_CP):

    assert len(GS6_CP.get_radius_of_gyration()) == 5
    assert abs(GS6_CP.get_radius_of_gyration()[0] - 5.728453763896514) < 0.0001
    assert abs(GS6_CP.get_radius_of_gyration(R1=1,R2=3)[0] - 2.8815625777553278) < 0.0001

def test_get_t(GS6_CP):

    assert len(GS6_CP.get_t()) == 5
    assert abs(GS6_CP.get_t()[0] - 0.30286034750848245) < 0.0001
    assert abs(GS6_CP.get_t(R1=1,R2=3)[0] - 0.0766270900235029) < 0.0001

def test_get_internal_scaling(GS6_CP):

    assert len(GS6_CP.get_internal_scaling()[0]) == 6
    assert abs(np.mean(GS6_CP.get_internal_scaling()[1][1]) - 3.784701967415692) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling()[1][2]) - 6.524909331438218) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling(R1=1,R2=4)[1][1]) - 3.7353248) < 0.001


def test_get_internal_scaling_RMS(GS6_CP):

    assert len(GS6_CP.get_internal_scaling_RMS()[0]) == 6
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS()[1][1]) - 3.793829294427209) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS()[1][2]) - 6.578195001335268) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS(R1=1,R2=4)[1][1]) - 3.7455646317574534) < 0.001


def test_get_internal_scaling_RMS(GS6_CP):

    assert len(GS6_CP.get_internal_scaling_RMS()[0]) == 6
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS()[1][1]) - 3.793829294427209) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS()[1][2]) - 6.578195001335268) < 0.0001
    assert abs(np.mean(GS6_CP.get_internal_scaling_RMS(R1=1,R2=4)[1][1]) - 3.7455646317574534) < 0.001


def test_soursop_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "soursop" in sys.modules


def test_DSSP(NTL9_CP):
    SS = NTL9_CP.get_secondary_structure_DSSP()
    print(SS[0])


def test_BBSEG(NTL9_CP):
    SS = NTL9_CP.get_secondary_structure_BBSEG()
    print(SS[0])



def test_phi_psi_bbseg(NTL9_CP):
    """
    Test that ensures our representation of the BBSEG2 matrix
    reproduces the version from CAMPARI

    """

    # read in BBSEG file
    bbseg_file = soursop.get_data('bbseg2.dat')
    with open(bbseg_file,'r') as fh:
        content = fh.readlines()

    idx=0
    for psi in range(179,-180, -10):

        local = ''
        for phi in range(-179,180, 10):
            local = local + '%i ' %(NTL9_CP._SSProtein__phi_psi_bbseg([phi],[psi])[0])

        assert(content[idx].strip() == local.strip())
        idx=idx+1



def test_get_distance_map(GS6_CO):
    for protein_index in range(len(GS6_CO.proteinTrajectoryList)):
        distance_map, stddev_map = GS6_CO.proteinTrajectoryList[0].get_distance_map()

        # verify that the shapes match
        assert distance_map.shape == stddev_map.shape

        # verify that we obtain an upper triangular matrix
        assert np.allclose(distance_map, np.triu(distance_map)) is True



def test_get_hydrodynamic_radius(GS6_CO):

    CP = GS6_CO.proteinTrajectoryList[0]
    rh = CP.get_hydrodynamic_radius()
    assert (11.840569781179006 - rh[0]) < 0.001


# ====
def test_init_from_trajectory(GS6_CO, NTL9_CO, GS6_CP, NTL9_CP):
    trajs = [GS6_CO, NTL9_CO]
    proteins = [GS6_CP, NTL9_CP]
    for ct_traj, ct_protein in zip(trajs, proteins):
        protein_from_md = ssprotein.SSProtein(ct_traj)
        assert protein_from_md.n_frames > 0
        assert protein_from_md.n_residues > 0
        assert protein_from_md.length() == ct_protein.length()

        protein_from_ct = ssprotein.SSProtein(ct_traj.traj)
        assert protein_from_ct.n_frames > 0
        assert protein_from_ct.n_residues > 0
        assert protein_from_ct.length() == ct_protein.length()


def test_init_from_trajectory_invalid():
    args = [1, 1.0, 'test', None]
    for arg in args:
        with pytest.raises(RuntimeError):
            protein = ssprotein.SSProtein(arg)


def test_init_from_trajectory_debug(GS6_CO, NTL9_CO, GS6_CP, NTL9_CP):
    trajs = [GS6_CO, NTL9_CO]
    proteins = [GS6_CP, NTL9_CP]
    for ct_traj, ct_protein in zip(trajs, proteins):
        protein_from_md = ssprotein.SSProtein(ct_traj.traj, debug=True)
        assert protein_from_md.n_frames > 0
        assert protein_from_md.n_residues > 0
        assert protein_from_md.length() == ct_protein.length()

        protein_from_ct = ssprotein.SSProtein(ct_traj, debug=True)
        assert protein_from_ct.n_frames > 0
        assert protein_from_ct.n_residues > 0
        assert protein_from_ct.length() == ct_protein.length()


def test_properties(GS6_CO, NTL9_CO, GS6_CP, NTL9_CP):
    properties = 'resid_with_CA,ncap,ccap,n_frames,n_residues,residue_index_list'.split(',')
    trajs = [GS6_CO, NTL9_CO]
    proteins = [GS6_CP, NTL9_CP]
    for ct_traj, ct_protein in zip(trajs, proteins):
        protein_from_ct = ssprotein.SSProtein(ct_traj)

        for prop in properties:
            assert getattr(protein_from_ct, prop) == getattr(ct_protein, prop)


def test_repr(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        repr_string = repr(protein)
        rebuilt_repr = "SSProtein (%s): %i res and %i frames" % (hex(id(protein)), protein.n_residues, protein.n_frames)
        assert rebuilt_repr == repr_string


def test_len(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        length = len(protein)
        attributes = protein.length()
        assert length == protein.n_frames
        assert attributes == (protein.n_residues, protein.n_frames)


def test_check_weights_invalid_weights_type(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = 'abcdefgh'
        with pytest.raises(ValueError):
            protein._SSProtein__check_weights(weights=weights)


def test_check_weights_valid_uniform_weights_length(GS6_CP, NTL9_CP):
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        protein._SSProtein__check_weights(weights=normalized_weights)


def test_check_weights_valid_uniform_weights_length_low_etol(GS6_CP, NTL9_CP):
    default_weight = 1.0
    tolerance = np.finfo(np.longdouble).precision
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        protein._SSProtein__check_weights(weights=normalized_weights, etol=tolerance)


def test_check_weights_valid_uniform_weights_length_large_etol(GS6_CP, NTL9_CP):
    default_weight = 1.0
    tolerance = 0.1
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        out = protein._SSProtein__check_weights(weights=normalized_weights, etol=tolerance)
        assert np.allclose(out, normalized_weights)


def test_check_weights_valid_nonuniform_weights_length_low_etol(GS6_CP, NTL9_CP):
    tolerance = np.finfo(np.longdouble).precision
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [w for w in range(1, protein.n_frames + 1)]
        normalized_weights = [w/sum(weights) for w in weights]
        protein._SSProtein__check_weights(weights=normalized_weights, etol=tolerance)


def test_check_weights_invalid_uniform_weights_length(GS6_CP, NTL9_CP):
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames - 1)]
        normalized_weights = [w/protein.n_frames for w in weights]
        with pytest.raises(SSException):
            protein._SSProtein__check_weights(weights=normalized_weights)


def test_check_weights_valid_uniform_weights_nondefault_stride(GS6_CP, NTL9_CP):
    stride = 2  # default stride is None
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        protein._SSProtein__check_weights(weights=normalized_weights, stride=stride)


def test_check_weights_short_uniform_weights_nondefault_stride(GS6_CP, NTL9_CP):
    stride = 2  # default stride is None
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(0, protein.n_frames, stride)]
        normalized_weights = [w/protein.n_frames for w in weights]
        with pytest.raises(SSException):
            protein._SSProtein__check_weights(weights=normalized_weights, stride=stride)


# == SSProtein.__get_first_and_last
# Should there be a test for R1 == R2?
def test_get_first_and_last_flip_residue_indices(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        residue_indices = list(range(protein.n_residues))
        mid_point = (protein.n_residues//2) - 1
        lower = residue_indices[:mid_point]
        upper = residue_indices[mid_point+1:]

        # Shuffle to ensure that the indices chosen, whilst random
        # are always where R1 > R2.
        random.shuffle(lower)
        random.shuffle(upper)
        r1 = random.choice(upper)
        r2 = random.choice(lower)
        protein._SSProtein__get_first_and_last(r1, r2)


# == SSProtein.__check_stride
def test_check_invalid_strides(GS6_CP, NTL9_CP):
    strides = [-1, 0, 100]  # none of the proteins have more than 100 frames.
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        for stride in strides:
            with pytest.raises(SSException):
                protein._SSProtein__check_stride(stride)


# == SSProtein.__check_single_residue
def test_check_invalid_single_residue(GS6_CP, NTL9_CP):
    invalid_residues = [-1, 100]  # none of the proteins have more than 100 residues
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        for invalid_residue_index in invalid_residues:
            with pytest.raises(SSException):
                protein._SSProtein__check_single_residue(invalid_residue_index)


# == SSProtein.__check_contains_CA
'''
def test_check_contains_CA_raises_exception(GS6_CP, NTL9_CP):
    # Requires modifying the protein topology to insert one or more fake residues where CA is missing
    # or replaced by something else.
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        residues_no_CA = protein.topology.select('resid %i to %i and atom not type "CA"' % (0, protein.n_residues - 1))
        print(residues_no_CA)
        for residue in residues_no_CA:
            with pytest.raises(SSException):
                protein._SSProtein__check_contains_CA(residue)
'''

def test_check_contains_CA_successful(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        residues_CA = protein.resid_with_CA
        for residue in residues_CA:
            return_value = protein._SSProtein__check_contains_CA(residue)
            assert return_value == None


# == SSProtein.__get_selection_atoms
def test_get_selection_atoms_region_of_size_2(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        region = (0, protein.n_residues - 1)
        choices = itertools.product([True, False], repeat=2)
        for backbone, heavy in choices:
            protein._SSProtein__get_selection_atoms(region, backbone, heavy)


def test_get_selection_atoms_invalid_region_of_size_3(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        region = (0, protein.n_residues - 1, 0)  # the last index doesn't matter (will raise exception regardless)
        choices = itertools.product([True, False], repeat=2)
        for backbone, heavy in choices:
            with pytest.raises(SSException):
                protein._SSProtein__get_selection_atoms(region, backbone, heavy)


# == SSProtein.get_amino_acid_sequence
def test_get_amino_acid_sequence(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        seq1 = protein.get_amino_acid_sequence(oneletter=True, numbered=False)

        # should there be a check for caps?
        assert len(seq1) == protein.n_residues

        seq2 = protein.get_amino_acid_sequence(oneletter=False, numbered=True)
        assert len(seq2) == protein.n_residues

        start_residue_number    = int(seq2[0].split('-')[-1])
        end_residue_number      = int(seq2[-1].split('-')[-1])

        assert start_residue_number == 1
        assert end_residue_number == len(seq2)
        assert (end_residue_number - start_residue_number) == len(seq2) - 1


def test_get_multiple_CA_index_existing(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        for residue_id in protein.resid_with_CA:
            atom_index = protein.get_multiple_CA_index(residue_id)

            # Validate the results and ensure they are ints
            assert len(atom_index) == 1
            assert residue_id in protein.resid_with_CA
            assert residue_id in protein._SSProtein__resid_with_CA
            assert type(atom_index[0]) in [np.int16, np.int32, np.int64]
            assert atom_index[0] > 0


def test_get_multiple_CA_index_invalid_residue_number(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        max_residue = protein.n_residues
        for residue_index in range(max_residue + 1, max_residue + protein.n_residues):
            with pytest.raises(SSException):
                protein.get_multiple_CA_index(residue_index)


def test_get_multiple_CA_index_invalid_residue_number_list(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        max_residue = protein.n_residues
        residue_list = list(range(max_residue + 1, max_residue + protein.n_residues))
        atoms_with_CA = protein.get_multiple_CA_index(resID_list=residue_list)

        assert type(atoms_with_CA) == list
        assert len(atoms_with_CA) == 0


def test_calculate_all_CA_distances_invalid_residue_number(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        max_residue = protein.n_residues
        for residue_index in range(max_residue + 1, max_residue + protein.n_residues):
            index = protein.calculate_all_CA_distances(residue_index)

            # Invalid indices return -1
            assert index == -1


# == SSProtein._SSProtein__residue_atom_com
def test_residue_atom_com_new_resid_atom_name_None(GS6_CO, NTL9_CO):
    trajs = [GS6_CO, NTL9_CO]
    for traj in trajs:
        # instantiate a new protein object since the previous reference is modified
        # elsewhere - hence our residue count will be off.
        protein = ssprotein.SSProtein(traj)
        unavailable_residue_index = protein.n_residues + 1
        protein.residue_atom_com(unavailable_residue_index, atom_name=None)
        assert len(protein._SSProtein__residue_atom_table) == protein.n_residues + 1


def test_residue_atom_com_new_resid_atom_name_CA(GS6_CO, NTL9_CO):
    trajs = [GS6_CO, NTL9_CO]
    for traj in trajs:
        # instantiate a new protein object since the previous reference is modified
        # elsewhere - hence our residue count will be off.
        protein = ssprotein.SSProtein(traj)
        unavailable_residue_index = protein.n_residues + 1
        protein.residue_atom_com(unavailable_residue_index, atom_name='CA')
        assert len(protein._SSProtein__residue_atom_table) == protein.n_residues + 1



def test_get_distance_map_weights(GS6_CP, NTL9_CP):
    default_weight = 1.0
    modes = 'CA,COM'.split(',')
    rms_options = [False, True]
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        for mode in modes:
            for rms_option in rms_options:
                results = protein.get_distance_map(mode=mode, RMS=rms_option, weights=normalized_weights)
                distance_map, std_dev = results

                protein_residues = len(protein.resid_with_CA)
                expected_shape = (protein_residues, protein_residues)

                assert distance_map.shape == expected_shape
                assert std_dev.shape == expected_shape

                if rms_option is True:
                    assert np.count_nonzero(np.tril(std_dev, -1)) == 0


def test_get_local_collapse_invalid_bins(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        # Wrong number of bins
        with pytest.raises(SSException):
            protein.get_local_collapse(bins=[1])

        # Wrong bins type
        with pytest.raises(SSException):
            protein.get_local_collapse(bins='a,b,c,d'.split(','))

        # Uneven bins spacing
        bins = [1,2,4,6]
        with pytest.raises(SSException):
            protein.get_local_collapse(bins=bins)


def test_get_local_collapse_invalid_window_size(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        number_of_residues = protein.n_residues + 1
        with pytest.raises(SSException):
            protein.get_local_collapse(window_size=number_of_residues)


def test_get_local_collapse_successful_run(NTL9_CP):
    # get_local_collapse(self, window_size=10, bins=None, verbose=True)
    window_size = 10
    meanData, stdData, histo, bins = NTL9_CP.get_local_collapse()

    expected_length = NTL9_CP.n_residues - (window_size - 1)
    assert len(meanData) == expected_length
    assert len(stdData) == expected_length
    assert len(histo) == expected_length
    assert type(bins) != None


# SSProtein.get_angle_decay
def test_get_angle_decay_return_all_pairs(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        result = protein.get_angle_decay(return_all_pairs=True)
        assert len(result) == 2

        return_matrix, all_pairs = result

        # check this has been calculated correctly  -we expect all_pairs
        # to be a dictionary that reports on every unique inter-residue pair
        assert len(all_pairs) == np.sum(list(range(len(return_matrix), 0, -1)))
        


def test_get_angle_decay_no_return_all_pairs(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        return_matrix = protein.get_angle_decay(return_all_pairs=False)


        num_caps = 0
        if protein.ncap:
            num_caps += 1
        if protein.ccap:
            num_caps += 1

        assert len(return_matrix) == protein.n_residues - num_caps


# SSProtein.get_contact_map
def test_get_contact_map_weights(GS6_CP, NTL9_CP):
    default_weight = 1.0
    modes = 'ca,closest,closest-heavy,sidechain'.split(',')
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        for mode in modes:
            results = protein.get_contact_map(mode=mode, weights=normalized_weights)  # the other options have been test
            contact_map, contact_map_order = results

            protein_residues = len(protein.resid_with_CA)
            expected_shape = (protein_residues, protein_residues)

            assert contact_map.shape == expected_shape
            assert len(contact_map_order.shape) == 1
            assert contact_map_order.shape[0] == protein_residues


def test_get_contact_map_weights_sidechain_heavy(GS6_CP, NTL9_CP):
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    mode = 'sidechain-heavy'
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]

        # note in a previous v
        protein.get_contact_map(mode=mode, weights=normalized_weights)  # the other options have been tested

    # this should fail with an SSException
    mode = 'invalid-mode'
    for protein in proteins:
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]


        with pytest.raises(SSException):
            protein.get_contact_map(mode=mode, weights=normalized_weights)  # the other options have been tested
        
        


# SSProtein.get_sidechain_alignment_angle
def test_get_sidechain_alignment_angle_unsupported_angle(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    R1 = 0
    R2 = 0
    for protein in proteins:
        with pytest.raises(SSException):
            protein.get_sidechain_alignment_angle(R1, R2, sidechain_atom_1='invalid_atom')

        with pytest.raises(SSException):
            protein.get_sidechain_alignment_angle(R1, R2, sidechain_atom_2='invalid_atom')


def test_get_sidechain_alignment_angle_successful_random(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in [NTL9_CP]:
        residue_indices = list(range(protein.n_residues))
        mid_point = (protein.n_residues//2) - 1
        lower = residue_indices[1:mid_point]
        upper = residue_indices[mid_point+1:-1]

        # Shuffle to ensure that the indices chosen, whilst random
        # are always where R1 > R2.
        random.shuffle(lower)
        random.shuffle(upper)
        r1 = random.choice(upper)
        r2 = random.choice(lower)

        alignment = protein.get_sidechain_alignment_angle(r1, r2)
        assert len(alignment) == protein.n_frames

    for protein in [GS6_CP]:
        residue_indices = list(range(protein.n_residues))
        mid_point = (protein.n_residues//2) - 1
        lower = residue_indices[1:mid_point]
        upper = residue_indices[mid_point+1:-1]

        # Shuffle to ensure that the indices chosen, whilst random
        # are always where R1 > R2.
        random.shuffle(lower)
        random.shuffle(upper)
        r1 = random.choice(upper)
        r2 = random.choice(lower)

        with pytest.raises(SSException):
            alignment = protein.get_sidechain_alignment_angle(r1, r2)
        


def test_get_sidechain_alignment_angle_invalid_r1(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        max_residue = protein.n_residues
        r1 = max_residue + protein.n_residues
        r2 = 0

        with pytest.raises(SSException):
            protein.get_sidechain_alignment_angle(r1, r2)


def test_get_sidechain_alignment_angle_invalid_r2(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    for protein in proteins:
        max_residue = protein.n_residues
        r1 = 0
        r2 = max_residue + protein.n_residues

        with pytest.raises(SSException):
            protein.get_sidechain_alignment_angle(r1, r2)


#get_local_to_global_correlation(self, mode='COM', n_cycles=100, max_num_pairs=10, stride=20, weights=False, verbose=True)
def test_get_local_to_global_correlation_unsupported_modes(GS6_CP, NTL9_CP):
    proteins = [GS6_CP, NTL9_CP]
    invalid_modes = 'unknown,unsupported'.split(',')
    for protein in proteins:
        for mode in invalid_modes:
            with pytest.raises(SSException):
                protein.get_local_to_global_correlation(mode=mode)


"""
# Requires additional work. Exception raised:
# "raise SSException('Something when wrong when comparing stride-derived Rg and internal distances, this is a bug in the code...)')"
def test_get_local_to_global_correlation_defaults(GS6_CP, NTL9_CP, cta_protein_helper):
    num_copies = 5
    proteins = [GS6_CP, NTL9_CP]
    modes = 'CA,COM'.split(',')
    for protein_obj in proteins:
        protein = cta_protein_helper.lengthen_protein_trajectory(protein_obj, num_copies)
        for mode in modes:
            results = protein.get_local_to_global_correlation(mode=mode)

            assert len(results) == 4
            for result in results:
                assert len(result) > 0
"""


def test_get_local_to_global_correlation_valid_weights(GS6_CP, NTL9_CP, cta_protein_helper):
    num_copies = 5
    stride = 1
    default_weight = 1.0
    proteins = [GS6_CP, NTL9_CP]
    modes = 'CA,COM'.split(',')
    for protein_obj in proteins:
        protein = cta_protein_helper.lengthen_protein_trajectory(protein_obj, num_copies)
        weights = [default_weight for frame in range(protein.n_frames)]
        normalized_weights = [w/protein.n_frames for w in weights]
        for mode in modes:
            results = protein.get_local_to_global_correlation(mode=mode, stride=stride, weights=normalized_weights)

            assert len(results) == 4
            for result in results:
                assert len(result) > 0


def test_get_regional_SASA(GS6_CP, NTL9_CP, cta_protein_helper):
    num_copies = 5
    proteins = [GS6_CP, NTL9_CP]
    for protein_obj in proteins:
        protein = cta_protein_helper.lengthen_protein_trajectory(protein_obj, num_copies)

        residue_indices = list(range(protein.n_residues))
        mid_point = (protein.n_residues//2) - 1
        lower = residue_indices[1:mid_point]
        upper = residue_indices[mid_point+1:-1]

        # Shuffle to ensure that the indices chosen, whilst random
        # are always where R1 > R2.
        random.shuffle(lower)
        random.shuffle(upper)
        r1 = random.choice(upper)
        r2 = random.choice(lower)

        rsasa = protein.get_regional_SASA(r1, r2)
        assert rsasa != None


def test_get_site_accessibility_resid(GS6_CP, NTL9_CP, cta_protein_helper):
    num_copies = 5
    proteins = [GS6_CP, NTL9_CP]
    for protein_obj in proteins:
        protein = cta_protein_helper.lengthen_protein_trajectory(protein_obj, num_copies)

        residue_indices = list(range(protein.n_residues))
        protein.get_site_accessibility(residue_indices, mode='resid')


def test_get_site_accessibility_residue_names(GS6_CP, NTL9_CP, cta_protein_helper):
    num_copies = 5
    proteins = [GS6_CP, NTL9_CP]
    residue_names = ['GLY']
    for protein_obj in proteins:
        protein = cta_protein_helper.lengthen_protein_trajectory(protein_obj, num_copies)

        protein.get_site_accessibility(residue_names, mode='residue_type')



import os
import pytest
import numpy as np
import soursop.ssnmr as nmr
from soursop.ssexceptions import SSException


def test_compute_random_coil_chemical_shifts_extreme_temperatures():
    seq = 'ARNDE'
    temperatures = [
        -273.15,  # absolute zero (celcius)
        -273,
        100.1,
        101
    ]

    for temperature in temperatures:
        with pytest.raises(SSException):
            nmr.compute_random_coil_chemical_shifts(seq, temperature=temperature, use_ggxgg=True, use_perdeuteration=False)


def test_compute_random_coil_chemical_shifts_pH_low():
    seq = 'ARNDE'
    temperature = 30
    pHs = [
        -1.0,
        -1,
        15.0,
        15
    ]

    for pH in pHs:
        with pytest.raises(SSException):
            nmr.compute_random_coil_chemical_shifts(seq, temperature=temperature, pH=pH, use_ggxgg=True, use_perdeuteration=False)


def test_round3_as_strings():
    numbers = [
        #-0.4956, 
        #-0.1,
        # 0, # This produces a strange output as a string: `0.0000.100`
        0.1, 
        0.2, 
        1/3, 
        0.5
    ]
    expected = [
        #'-0.496', 
        #'-0.100',
        # '0.000', 
        '0.100', 
        '0.200', 
        '0.333', 
        '0.500'
    ]

    for num, exp_num in zip(numbers, expected):
        num_str = nmr.__round3(num, asFloat=False)
        assert type(num_str) is str
        assert num_str == exp_num



"""
Unit and regression test for the sstrajectory module.
"""
# Import package, test suite, and other packages as needed
import soursop
import hashlib
from soursop import sstrajectory
from soursop.ssexceptions import SSException
from pathlib import Path
import pytest
import sys
import random
import itertools
import platform
import tempfile
import numpy as np
import os
from soursop.configs import TMP_DIR


from . import conftest
#import conftest

# -------------------------------------------------------------------------------------------------
# SSTrajectory `__init__` tests.
#
# Also implicitly tests:
#       `__readTrajectory`
#       `__get_proteins`
#       `__get_proteins_by_residue`

def test_trajectory_repr_string(GS6_CO):
    repr_string = (hex(id(GS6_CO)), GS6_CO.num_proteins, GS6_CO.n_frames)
    return repr_string == repr(GS6_CO)


def test_trajectory_len(GS6_CO):
    assert len(GS6_CO) == len(GS6_CO.traj)
    assert len(GS6_CO) == GS6_CO.n_frames


def test_read_in_no_trajectory_and_no_topology():
    with pytest.raises(SSException):
        sstrajectory.SSTrajectory()


def test_read_in_no_trajectory_and_no_topology_but_use_custom_trajectory(GS6_CO):
    trajectory = GS6_CO.traj
    sstrajectory.SSTrajectory(TRJ=trajectory)


def test_read_in_no_trajectory():
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6.pdb')
    with pytest.raises(SSException):
        sstrajectory.SSTrajectory(pdb_filename=pdb_filename)


def test_read_in_trajectory_pdblead(GS6_CO):
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    traj = sstrajectory.SSTrajectory(pdb_filename=pdb_filename, trajectory_filename=traj_filename, pdblead=True)

    # Since we're using the PDB as an initial frame, the number of frames should have increased by 1.
    assert len(traj) == len(GS6_CO) + 1


def test_trajectory_initialization_debug():
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    sstrajectory.SSTrajectory(pdb_filename=pdb_filename, trajectory_filename=traj_filename, debug=True)


def test_read_in_no_topology_xtc():
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    with pytest.raises(SSException):
        sstrajectory.SSTrajectory(trajectory_filename=traj_filename)


def test_read_in_no_topology_dcd():
    """The `SSTrajectory.__init__` references that `.dcd` files are supported too."""
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.dcd')
    with pytest.raises(SSException):
        sstrajectory.SSTrajectory(trajectory_filename=traj_filename)


def test_read_in_compare_trajectories(GS6_CO):
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename, pdb_filename=pdb_filename)

    assert np.allclose(trajectory.traj.xyz, GS6_CO.traj.xyz)


def test_read_in_protein_grouping_simple():
    protein_groups = [[0]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    # verify that we have loaded the number of proteins expected
    assert trajectory.num_proteins == len(protein_groups)


def test_read_in_protein_grouping_multiple():
    protein_groups = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    # verify that we have loaded the number of proteins expected
    assert trajectory.num_proteins == len(protein_groups)


def test_read_in_protein_grouping_invalid_residues():
    protein_groups = [[1000, 1001, 1002], [1003, 1004, 1005], [1006, 1007, 1008]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    # Since this is a failing but non-disruptive test (i.e. no Exceptions), we
    # check the number of proteins which should be 0.
    assert trajectory.num_proteins == 0


def test_read_in_protein_grouping_multiple_mixed_order():
    # TODO: This test passes - it shouldn't. Look into this further.
    protein_groups = [[1, 2, 0], [10, 3, 7], [11, 1, 3]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    # verify that we have loaded the number of proteins expected
    assert trajectory.num_proteins == len(protein_groups)


# -------------------------------------------------------------------------------------------------


def test_get_intra_chain_distance_map_zero(GS6_CO):
    # TODO: This returns an upper triangle copy of the original distance map. Investigate further.
    distance_map_zero, stddev_map_zero = GS6_CO.get_interchain_distance_map(0, 0)
    distance_map, stddev_map = GS6_CO.proteinTrajectoryList[0].get_distance_map()

    assert np.allclose(np.triu(distance_map_zero), distance_map)
    assert np.allclose(np.triu(stddev_map_zero), stddev_map, atol=1e-6)


def test_get_intra_chain_distance_map_zero_using_center_of_mass(GS6_CO):
    # TODO: This returns an upper triangle copy of the original distance map. Investigate further.
    distance_map_zero, stddev_map_zero = GS6_CO.get_interchain_distance_map(0, 0, mode='COM')
    distance_map, stddev_map = GS6_CO.proteinTrajectoryList[0].get_distance_map(mode='COM')

    assert np.allclose(np.triu(distance_map_zero), distance_map)
    assert np.allclose(np.triu(stddev_map_zero), stddev_map, atol=1e-6)


def test_get_intra_chain_distance_map_protein_groups():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    for protein_group in itertools.combinations(range(len(protein_groups)), r=2):
        protein_a, protein_b = protein_group
        distance_map, stddev_map = trajectory.get_interchain_distance_map(protein_a, protein_b)
        assert distance_map.shape == stddev_map.shape
        assert np.count_nonzero(distance_map) == len(distance_map.flatten())  # since the residue indices are unique


def test_get_intra_chain_distance_map_protein_groups_with_residue_indices():
    # Note that the residues for the resID1 and resID2 indices are with respect to the residues of the protein chain
    # and not the full protein itself.
    protein_groups_residues = [[0, 1, 2, 3, 4, 5, 6, 7, 8],
                               [10, 11, 12, 13, 14, 15, 16, 17, 18],
                               [20, 21, 22, 23, 24, 25, 26, 27, 28],
                               [30, 31, 32, 33, 34, 35, 36, 37, 38],
                               [40, 41, 42, 43, 44, 45, 46, 47, 48]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups_residues)

    # select pairwise protein ids (no repetitions).
    for protein_group in itertools.combinations(range(len(protein_groups_residues)), r=2):
        protein_a, protein_b = protein_group

        # now, let's randomly choose a set of residues to select for `protein_a` and `protein_b`
        num_residues_a = random.randint(1, len(protein_groups_residues[protein_a]))
        num_residues_b = random.randint(1, len(protein_groups_residues[protein_b]))

        protein_a_residues = protein_groups_residues[protein_a]
        a_residues = list(sorted(random.sample(protein_a_residues, num_residues_a)))
        a_residues = [(r - protein_a_residues[0]) for r in a_residues]  # offset by the starting residue number

        protein_b_residues = protein_groups_residues[protein_b]
        b_residues = list(sorted(random.sample(protein_b_residues, num_residues_b)))
        b_residues = [(r - protein_b_residues[0]) for r in b_residues]  # offset by the starting residue number

        distance_map, stddev_map = trajectory.get_interchain_distance_map(protein_a, protein_b)
        assert distance_map.shape == stddev_map.shape


def test_export_intra_chain_distance_map_protein_groups_with_residue_indices():
    # Note that the residues for the resID1 and resID2 indices are with respect to the residues of the protein chain
    # and not the full protein itself.
    protein_groups_residues = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                               [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                               [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
                               [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
                               [40, 41, 42, 43, 44, 45, 46, 47, 48, 49]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups_residues)

    # select pairwise protein ids (no repetitions).
    for protein_index, protein_group in enumerate(itertools.combinations(range(len(protein_groups_residues)), r=2)):
        protein_a, protein_b = protein_group

        # now, let's randomly choose a set of residues to select for `protein_a` and `protein_b`
        num_residues_a = random.randint(1, len(protein_groups_residues[protein_a]))
        num_residues_b = random.randint(1, len(protein_groups_residues[protein_b]))

        protein_a_residues = protein_groups_residues[protein_a]
        a_residues = list(sorted(random.sample(protein_a_residues, num_residues_a)))
        a_residues = [(r - protein_a_residues[0]) for r in a_residues]  # offset by the starting residue number

        protein_b_residues = protein_groups_residues[protein_b]
        b_residues = list(sorted(random.sample(protein_b_residues, num_residues_b)))
        b_residues = [(r - protein_b_residues[0]) for r in b_residues]  # offset by the starting residue number

        distance_map, stddev_map = trajectory.get_interchain_distance_map(protein_a, protein_b)
        assert distance_map.shape == stddev_map.shape

        temp_save_filepath = os.path.join(TMP_DIR, 'gs6_map_{index}_{protein}'.format(index=protein_index,
                                                                                      protein=protein_index))
        temp_filename = '{filename}.csv'.format(filename=temp_save_filepath)


def test_intrachain_inter_residue_atomic_distance():

    # A custom version of NTL9 is needed, since we want to have multiple protein chains
    protein_groups_residues = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                               [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                               [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
                               [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
                               [40, 41, 42, 43, 44, 45, 46, 47, 48, 49]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups_residues)

    for protein_group in itertools.combinations(range(len(protein_groups_residues)), r=2):
        protein_a, protein_b = protein_group
        a_residues = protein_groups_residues[protein_a]
        b_residues = protein_groups_residues[protein_b]

        residue_a = random.randint(0, len(a_residues) - 1)
        residue_b = random.randint(0, len(b_residues) - 1)

        distances = trajectory.get_interchain_distance(protein_a, protein_b, residue_a, residue_b)
        assert len(distances) == len(a_residues)


def test_intrachain_inter_residue_atomic_distance_by_name():
    """
    # TODO: Revise.
    # A custom version of NTL9 is needed, since we want to have multiple protein chains
    protein_groups_residues = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                               [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                               [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
                               [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
                               [40, 41, 42, 43, 44, 45, 46, 47, 48, 49]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups_residues)

    atom_names = dict()
    for residue in trajectory.traj.topology.residues:
        atom_names[residue.index] = [atom for atom in residue.atoms]#[atom.name for atom in residue.atoms]

    for protein_group in itertools.combinations(range(len(protein_groups_residues)), r=2):
        protein_a, protein_b = protein_group
        a_residues = protein_groups_residues[protein_a]
        b_residues = protein_groups_residues[protein_b]

        residue_a = random.randint(0, len(a_residues) - 1)
        residue_a_atoms = atom_names[protein_groups_residues[protein_a][residue_a]]
        residue_a_atom = random.choice(residue_a_atoms)

        residue_b = random.randint(0, len(b_residues) - 1)
        residue_b_atoms = atom_names[protein_groups_residues[protein_b][residue_b]]
        residue_b_atom = random.choice(residue_b_atoms)

        distances = trajectory.get_intrachain_interResidue_atomic_distance(protein_a, protein_b,
                                                                           residue_a, residue_b,
                                                                           A1=residue_a_atom, A2=residue_a_atom)
        assert len(distances) == len(a_residues)
    """
    pass


def test_get_interchain_distance():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    for protein_group in itertools.combinations(range(len(protein_groups)), r=2):
        protein_a, protein_b = protein_group

        # There appears to be an issue where R1 > R2 - the calculation crashes for sidechain-heavy.
        R1 = 0 #random.choice(range(len(protein_groups[protein_a])))
        R2 = 1 #random.choice(range(len(protein_groups[protein_b])))
        A1 = 'CA'
        A2 = 'CA'
        modes = 'atom,ca,closest,closest-heavy,sidechain,sidechain-heavy'.split(',')
        for mode in modes:
            distances = trajectory.get_interchain_distance(protein_a, protein_b, R1, R2, A1, A2, mode)
            assert len(distances) > 0


def test_get_interchain_distance_invalid_mode():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    mode = 'unknown'
    for protein_group in itertools.combinations(range(len(protein_groups)), r=2):
        protein_a, protein_b = protein_group

        # There appears to be an issue where R1 > R2 - the calculation crashes for sidechain-heavy.
        R1 = 0 #random.choice(range(len(protein_groups[protein_a])))
        R2 = 1 #random.choice(range(len(protein_groups[protein_b])))
        A1 = 'CA'
        A2 = 'CA'
        
        with pytest.raises(SSException):
            trajectory.get_interchain_distance(protein_a, protein_b, R1, R2, A1, A2, mode)


def test_get_interchain_distance_failing_atom():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    mode = 'atom'
    atoms1 = ['X', 'CA']
    atoms2 = atoms1[::-1]
    for protein_group in itertools.combinations(range(len(protein_groups)), r=2):
        protein_a, protein_b = protein_group

        # There appears to be an issue where R1 > R2 - the calculation crashes for sidechain-heavy.
        # See `test_get_interchain_distance`
        R1 = 0 #random.choice(range(len(protein_groups[protein_a])))
        R2 = 1 #random.choice(range(len(protein_groups[protein_b])))

        for A1, A2 in zip(atoms1, atoms2):
            with pytest.raises(SSException):
                trajectory.get_interchain_distance(protein_a, protein_b, R1, R2, A1, A2, mode)


def test_get_interchain_distance_failing_residues():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    residues1 = [0, 100]
    residues2 = residues1[::-1]
    A1 = 'CA'
    A2 = 'CA'
    modes = 'atom,ca,closest,closest-heavy,sidechain,sidechain-heavy'.split(',')
    
    for protein_group in itertools.combinations(range(len(protein_groups)), r=2):
        protein_a, protein_b = protein_group
        for mode in modes:
            for R1, R2 in zip(residues1, residues2):
                with pytest.raises(SSException):
                    trajectory.get_interchain_distance(protein_a, protein_b, R1, R2, A1, A2, mode)


def test_get_interchain_distance_invalid_protein_ids():
    protein_groups = [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]]
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'ntl9.xtc')  # ntl9 has 56 residues
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename,
                                           pdb_filename=pdb_filename,
                                           protein_grouping=protein_groups)

    R1 = 0
    R2 = 1
    A1 = 'CA'
    A2 = 'CA'
    modes = 'atom,ca,closest,closest-heavy,sidechain,sidechain-heavy'.split(',')
    
    protein_a = 100
    protein_b = 101
    for mode in modes:
        with pytest.raises(SSException):
            trajectory.get_interchain_distance(protein_a, protein_b, R1, R2, A1, A2, mode)


# -------------------------------------------------------------------------------------------------


def test_get_overall_hydrodynamic_radius(GS6_CO, NTL9_CO):
    gs6_radius = GS6_CO.get_overall_hydrodynamic_radius()
    ntl9_radius = NTL9_CO.get_overall_hydrodynamic_radius()
    assert len(gs6_radius) == GS6_CO.n_frames
    assert len(ntl9_radius) == NTL9_CO.n_frames


def test_get_overall_radius_of_gyration(GS6_CO, NTL9_CO):
    gs6_gyration = GS6_CO.get_overall_radius_of_gyration()
    ntl9_gyration = NTL9_CO.get_overall_radius_of_gyration()
    assert len(gs6_gyration) == GS6_CO.n_frames
    assert len(ntl9_gyration) == NTL9_CO.n_frames


def test_get_overall_asphericity(GS6_CO, NTL9_CO):
    gs6_asphericity = GS6_CO.get_overall_asphericity()
    ntl9_asphericity = NTL9_CO.get_overall_asphericity()
    assert len(gs6_asphericity) == GS6_CO.n_frames
    assert len(ntl9_asphericity) == NTL9_CO.n_frames


# -------------------------------------------------------------------------------------------------

def test_read_in_trajectory(GS6_CO):        

    # ensure 
    assert len(GS6_CO.traj) == 5
    assert len(GS6_CO.proteinTrajectoryList) == 1

"""
def test_protein_identification():
    pdb_filename = os.path.join(soursop.get_data('test_data'), 'gs6_invalid_r1.pdb')
    traj_filename = os.path.join(soursop.get_data('test_data'), 'gs6.xtc')
    trajectory = sstrajectory.SSTrajectory(trajectory_filename=traj_filename, pdb_filename=pdb_filename, debug=True)
    #print(trajectory)





"""


def test_get_overal_rh(GS6_CO):
    assert np.allclose(GS6_CO.get_overall_hydrodynamic_radius(), GS6_CO.proteinTrajectoryList[0].get_hydrodynamic_radius())

def test_get_overal_asphericity(GS6_CO):
    assert np.allclose(GS6_CO.get_overall_asphericity(), GS6_CO.proteinTrajectoryList[0].get_asphericity())

def test_get_overal_rg(GS6_CO):
    assert np.allclose(GS6_CO.get_overall_radius_of_gyration(), GS6_CO.proteinTrajectoryList[0].get_radius_of_gyration())

def test_get_overal_rg(GMX_2CHAINS):
    # compares rg of 2 chains vs. same values calculated by VMD

    # rg calculated in vmd
    vmd_rg = np.array([24.497053146362305, 24.938467025756836, 25.010461807250977, 27.36289405822754, 26.659685134887695, 24.56108283996582, 24.705211639404297, 27.144346237182617, 25.043672561645508, 27.67377471923828, 48.036869049072266, 23.930212020874023, 24.449384689331055, 24.491701126098633, 25.248193740844727, 27.898420333862305, 47.11823272705078, 47.213069915771484, 49.21323776245117, 27.429519653320312])

    assert np.allclose(GMX_2CHAINS.get_overall_radius_of_gyration(), vmd_rg)


def test_get_overal_asphericity(GMX_2CHAINS):
    # compares rg of 2 chains vs. same values calculated by VMD

    asph = np.array([0.36274249, 0.4079101, 0.5837932, 0.3413606, 0.30022731, 0.25667767, 0.24720234, 0.32076792, 0.26974903, 0.21184996, 0.62855842, 0.23021391, 0.16662495, 0.45061943, 0.54406795, 0.4482196, 0.53423135, 0.73352191, 0.75562579, 0.32473148])

    assert np.allclose(GMX_2CHAINS.get_overall_asphericity(), asph)

def test_get_overal_rh(GMX_2CHAINS):
    # compares rg of 2 chains vs. same values calculated by VMD

    asph = np.array([26.99809304, 27.25590015, 27.29754026, 28.59875282, 28.22148641, 27.03576021, 27.12021072, 28.48253016, 27.31671616, 28.76251356, 36.57647347, 26.66061797, 26.96999394, 26.99493729, 27.43425109, 28.87969746, 36.31506845, 36.3423472, 36.90213999, 28.63399989])

    assert np.allclose(GMX_2CHAINS.get_overall_hydrodynamic_radius(), asph)



import os
import sys
import pytest
import soursop
from soursop import sstrajectory


#NAMES = ['ACE_NH2', 'ACE_NME', 'ACE_UCAP', 'FOR_NH2',  'FOR_NME',  'FOR_UCAP', 'UCAP_NH2', 'UCAP_NME',  'UCAP_UCAP', 'ACE_NME_start_at_5', 'ACE_NME_multichain']

#CAPNAMES = ['ACE_NME_multichain']

#CCAP = ['ACE_NH2', 'ACE_NME', 'FOR_NH2',  'FOR_NME', 'UCAP_NH2', 'UCAP_NME', 'ACE_NME_start_at_5', 'ACE_NME_multichain']
#NCAP = ['ACE_NH2', 'ACE_NME', 'ACE_UCAP', 'FOR_NH2',  'FOR_NME',  'FOR_UCAP', 'ACE_NME_start_at_5', 'ACE_NME_multichain']




def test_cap_DSSP():
    """
    Tests DSSP (and hence R1/R2 selection) works
    """

    test_data_dir = soursop.get_data('test_data')
    pdb_files = ['ACE_NME_multichain', 'ACE_NME_start_at_5']
    for cap_name in pdb_files:
        cap_path = os.path.join(test_data_dir, 'cap_tests', '{}.pdb'.format(cap_name))
        cap_trajectory = sstrajectory.SSTrajectory(cap_path, cap_path)
        # print(cap_trajectory)
        # for cap_protein in cap_trajectory.proteinTrajectoryList:
        #     res_string = ''
        #     for i in cap_protein.topology.residues:
        #         #res_string = res_string + " %s " %(str(i.index))
        #         print(i)

        #     #atom_string = ''
        #     #for i in CP.topology.atoms:
        #     #    atom_string = atom_string + " " + str(i)

        #     print(cap_protein)
        #     print(res_string)
        #     print("Residue 1: %s" %(cap_protein.topology.select('residue 1')))
        #     print("Resid 1: %s" %(cap_protein.topology.select('resid 1')))
            
        #     # print(CP.atom_offset)
        #     # print(atom_string)
            




"""
soursop
Analysis package for all-atom simulations of proteins, with a specific focus on intrinsically disordered proteins.
"""
import setuptools
from setuptools import setup
import versioneer

short_description = __doc__.split("\n")

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except:
    long_description = "\n".join(short_description[2:]),


setup(
    # Self-descriptive entries which should always be present
    name='soursop',
    author='Alex Holehouse',
    author_email='alex.holehouse@wustl.edu',
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    #scripts=['scripts/<something>'],
    license='LGPLv3',
    url='https://github.com/holehouse-lab/soursop',

    # Which Python importable modules should be included when your package is installed
    packages=['soursop', "soursop.tests"],

    # Optional include package data to ship with your package
    # Comment out this line to prevent the files from being packaged with your software
    # Extend/modify the list to include/exclude other items as need be
    # > line below modified from default

    # line here means we include the files defined in MANIFEST.in - the package_data
    # line below is not sufficient to do this
    include_package_data=True,
    package_data={'soursop': ['data/'], 'soursop': ['data/test_data/']},

    # dependencies soursop requires
    install_requires=[
        "numpy>=1.20.0",
        "scipy>=1.5.0",
        "cython",
        "mdtraj>=1.9.5",
        "pandas>=0.23.0",
        "threadpoolctl>=2.2.0",
    ],


    # Additional entries you may want simply uncomment the lines you want and fill in the data
    #author_email='alex.holehouse@wustl.edu',      # Author email
    #url='http://www.my_package.com',              # Website

    # platforms=['Linux',
    #            'Mac OS-X',
    #            'Unix',
    #            'Windows'],            # Valid platforms your code works on, adjust to your flavor
    # python_requires=">=3.5",          # Python version restrictions

    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,

)



"""
camparitraj module

"""
##
##                                       _ _              _ 
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/ 
##
## Alex Holehouse (Pappu Lab and Holehouse Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##

from ._version import get_versions


def version():
    """
    Prints the full version and the git hash of the current commit

    Returns
    --------
    None 
        But prints version and git revision 
    """


    versions = get_versions()
    __version__ = versions['version']
    __git_revision__ = versions['full-revisionid']
    
    print(__version__)
    print(__git_revision__)


def version_full():
    """
    Returns the full camparitraj version as a string

    Returns
    --------
    str 
        camparitraj version
    """
    versions = get_versions()
    return versions['version']


def version_git_revision():
    """
    Returns the full git revision id as a string

    Returns
    --------
    str 
        camparitraj git revision id
    """

    versions = get_versions()
    return versions['full-revisionid']
    

if __name__ == "__main__":
    # Do something if this file is invoked on its own - e.g. print the version.
    print(version_full())



"""
camparitraj __ini__ file
A short description of the project.
"""
import os

# Add imports here
from .soursop import *

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions


# code that allows access to the data directory
_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_data(path):
    return os.path.join(_ROOT, 'data', path)


def get_version():
    return "%s - %s" % (str(__version__), str(__git_revision__))



"""
cttools contains misc. numerical python functions which provide some useful functionality 

"""
##
##                                       _ _              _ 
##   ___ __ _ _ __ ___  _ __   __ _ _ __(_) |_ _ __ __ _ (_)
##  / __/ _` | '_ ` _ \| '_ \ / _` | '__| | __| '__/ _` || |
## | (_| (_| | | | | | | |_) | (_| | |  | | |_| | | (_| || |
##  \___\__,_|_| |_| |_| .__/ \__,_|_|  |_|\__|_|  \__,_|/ |
##                     |_|                             |__/ 
##
## Alex Holehouse (Pappu Lab and Holehouse Lab)
## Simulation analysis package
## Copyright 2014 - 2021
##

import numpy as np

# ........................................................................
#
def chunks(l, n):
    """Yield successive n-sized chunks from l.

    Parameters
    ----------

    l : list
        The list from which chunks of size `n` will be selected.

    n : int
        The size of the chunks to select from an input list, `l`.
    """


    maxval = int(round(len(l) - len(l)%n))

    for i in range(0, maxval, int(n)):
        yield l[i:i + n]


# ........................................................................
#
def fix_histadine_name(name):
    """
    Corrects the histadine residue name which can be HIE, HID 
    or HIP and unifies all of these to HIS.

    Parameters
    ----------

    name : str
        The input residue name (3-letter code) that may be transformed if it is Histidine-related.

    Returns
    -------
    The transformed name of the Histidine residue, otherwise the input residue is returned.
    """
    
    if name == 'HIE' or name == 'HID' or name == 'HIP':
        return 'HIS'
    else:
        return name


# ........................................................................
#
def find_nearest(array, target):
    """
    Find the value nearest to a target in an array, returns a tuple
    with the value and the tuple.

    Parameters
    ----------

    array : np.array
        The array that will be searched for a target value.

    target: a value with dtype of `array`
        The search value to locate within `array`.

    Returns
    -------

    The first index and value of the target located within the input array.

    """
    idx = (np.abs(array-target)).argmin()
    return (idx,array[idx])


# ........................................................................
#
def powermodel(X, nu, R0):            
    """
    Function that computes the resulting power-model values.

    Parameters
    ----------

    X : array or numeric value
        An array or value whose power-model will be calculated.

    nu : int or float
        The exponent which the input array is raised to.

    R0: int or float
        A scaling constant.

    Returns
    -------
    array or numeric value

    """
    return R0*np.power(X,nu)            



import soursop
from soursop import sstrajectory
import pytest
import sys

test_data_dir = soursop.get_data('test_data')

CAPNAMES = ['ACE_NH2', 'ACE_NME', 'ACE_UCAP', 'FOR_NH2',  'FOR_NME',  'FOR_UCAP', 'UCAP_NH2', 'UCAP_NME',  'UCAP_UCAP', 'ACE_NME_start_at_5', 'ACE_NME_multichain']

#CAPNAMES = ['ACE_NME_multichain']

CCAP = ['ACE_NH2', 'ACE_NME', 'FOR_NH2',  'FOR_NME', 'UCAP_NH2', 'UCAP_NME', 'ACE_NME_start_at_5', 'ACE_NME_multichain']
NCAP = ['ACE_NH2', 'ACE_NME', 'ACE_UCAP', 'FOR_NH2',  'FOR_NME',  'FOR_UCAP', 'ACE_NME_start_at_5', 'ACE_NME_multichain']




def test_cap_DSSP():
    """
    Tests DSSP (and hence R1/R2 selection) works
    """
            
    for CN in CAPNAMES:
        CT = sstrajectory.SSTrajectory("%s/cap_tests/%s.pdb"%(test_data_dir, CN), "%s/cap_tests/%s.pdb"%(test_data_dir, CN))
        for CP in CT.proteinTrajectoryList:  
            print("")
            print(CP.get_amino_acid_sequence())
            assert len(CP.get_secondary_structure_DSSP()[0]) == 6
            
                

def test_cap_BBSEG():
    """
    Tests DSSP (and hence R1/R2 selection) works
    """
            
    for CN in CAPNAMES:
        CT = sstrajectory.SSTrajectory("%s/cap_tests/%s.pdb"%(test_data_dir, CN), "%s/cap_tests/%s.pdb"%(test_data_dir, CN))
        for CP in CT.proteinTrajectoryList:
            assert len(CP.get_secondary_structure_DSSP()[0]) == 6
                


def test_cap():
    """
    Test caps assignment works
    """
            
    for CN in CAPNAMES:
        CT = sstrajectory.SSTrajectory("%s/cap_tests/%s.pdb"%(test_data_dir, CN), "%s/cap_tests/%s.pdb"%(test_data_dir, CN))
        for CP in CT.proteinTrajectoryList:
            if CP.ccap:
                assert CN in CCAP
            if CP.ncap:
                assert CN in NCAP

